# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark online serving throughput.

Adapted from https://github.com/tyler-griggs/melange-release/blob/main/melange/profiling/gpu-benchmark.py, which is adapted from
https://github.com/vllm-project/vllm/blob/main/benchmarks/benchmark_serving.py

"""

import argparse
import asyncio
import json
import random
import sys
import time
from datetime import datetime, timezone
from typing import AsyncGenerator, List, Literal, Optional, Tuple

import aiohttp
import numpy as np

# (prompt len, output len, request latency)
REQUEST_LATENCY: List[Tuple[int, int, float]] = []
# (prompt len, output len, [per-token latencies])
TOKEN_LATENCY: List[Tuple[int, int, List[float]]] = []
TIME_TO_FIRST_TOKEN: List[float] = []
TEMPERATURE = 0.0


def print_err(
    *values: object,
    sep: str | None = " ",
    end: str | None = "\n",
    flush: Literal[False] = False,
) -> None:
    print(*values, sep=sep, end=end, file=sys.stderr, flush=flush)


def sample_requests(
    num_requests: int,
    config_input_len: int,
    config_output_len: int,
    workload_dataset_file: Optional[str] = None,
) -> List[Tuple[str, int, int, float]]:
    """Sample requests from prompt dataset or generate synthetic ones."""
    if workload_dataset_file:
        try:
            with open(workload_dataset_file) as f:
                data = json.load(f)
                # Return timestamp and request tuples
                requests = []
                for i, entry in enumerate(data):
                    # Limit the number of requests read from workload

                    # print(f"Request {i}: {entry}")
                    cur_timestamp = entry["Timestamp"]
                    next_timestamp = (
                        data[i + 1]["Timestamp"] if i < len(data) - 1 else cur_timestamp
                    )
                    interval = (next_timestamp - cur_timestamp) / 1000.0
                    for i, req in enumerate(entry["Requests"]):
                        requests.append(
                            (
                                req["Prompt"],
                                req["Prompt Length"],
                                req["Output Length"],
                                interval if i == len(entry["Requests"]) - 1 else 0,
                            )
                        )
                        if num_requests > 0 and len(requests) >= num_requests:
                            return requests
                # print('total requests: ', len(requests))
                # print('the least requests: ', requests[len(requests) - 1])
                return requests
        except Exception as e:
            print_err(f"Warning: Failed to load prompt dataset ({e})")
            return []
    else:
        # Original synthetic prompt generation
        requests = []
        for _ in range(num_requests):
            synthetic_prompt = "hi " * config_input_len
            # assign timestamp to -1 for all requests
            requests.append((synthetic_prompt, config_input_len, config_output_len, -1))
        return requests


async def get_request(
    input_requests: List[Tuple[str, int, int, float]],
    request_rate: float,
    num_requests: int,
    verbose: bool,
    use_workload_interval: bool = False,
) -> AsyncGenerator[Tuple[str, int, int, float], None]:
    requests = iter(input_requests)
    start_time = time.perf_counter()

    batch = 0
    for i, (prompt, prompt_len, output_len, interval) in enumerate(requests):
        current_time = time.perf_counter() - start_time
        if use_workload_interval:
            if verbose:
                print(
                    f"Batch {batch}, Request {i}: Sending at {current_time:.3f}s with interval {interval:.3f}s"
                )
            yield (prompt, prompt_len, output_len, interval)
            if interval > 0:
                batch += 1
                await asyncio.sleep(interval)
            continue
        else:
            interval = 0.0
            if i < num_requests - 1 and request_rate != float("inf"):
                # Sample the request interval from the exponential distribution.
                interval = np.random.exponential(1.0 / request_rate)
                if verbose:
                    print(
                        f"Request {i}: Generated exponential interval of {interval:.3f}s"
                    )

            request_with_next = (prompt, prompt_len, output_len, interval)
            if verbose:
                print(f"Request {i}: Sending at {current_time:.3f}s")
            yield request_with_next

            if request_rate == float("inf"):
                # If the request rate is infinity, then we don't need to wait.
                continue

            # The next request will be sent after the interval.
            await asyncio.sleep(interval)


def load_response(resp: str):
    return json.loads(resp)


async def send_request(
    idx: int,
    backend: str,
    api_url: str,
    api_key: Optional[str],
    model: str,
    prompt: str,
    prompt_len: int,
    output_len: int,
    next_in: float,
    best_of: int,
    use_beam_search: bool,
    stream: bool,
    verbose: bool,
    trace: bool,
) -> None:
    headers = {
        "User-Agent": "Benchmark Client",
    }
    if api_key is not None or api_key != "":
        headers["Authorization"] = f"Bearer {api_key}"

    streaming = stream
    if backend == "vllm":
        pload = {
            "model": model,
            "prompt": prompt,
            # "n": 1,
            # "best_of": best_of,
            # "use_beam_search": use_beam_search,
            "temperature": 0.0 if use_beam_search else TEMPERATURE,
            # "top_p": 1.0,
            "max_tokens": output_len,
            # "ignore_eos": True,
        }
        if stream:
            pload["stream"] = 1
        # Only apply "next_in" for simulator which requires no api_key.
        if next_in > 0.0 and (api_key is None or api_key == ""):
            pload["next_in"] = next_in
    else:
        raise ValueError(f"Unknown backend: {backend}")

    request_start_time = time.perf_counter()
    ts = datetime.now(timezone.utc)
    timeout = aiohttp.ClientTimeout(total=3 * 3600)
    status_code = None
    async with aiohttp.ClientSession(timeout=timeout) as session:
        while True:
            # print(f"Sending request: {api_url}:{pload}")
            async with session.post(api_url, headers=headers, json=pload) as response:
                status_code = response.status
                chunks = []
                token_latencies = []
                previous_token_time = time.perf_counter()
                first = True
                try:
                    if streaming:
                        async for chunk, _ in response.content.iter_chunks():
                            # Stream on: Each chunk in the response is the full response so far
                            chunks = [chunk]

                            now_time = time.perf_counter()
                            if first:
                                time_to_first = now_time - previous_token_time
                                first = False
                            else:
                                token_latencies.append(now_time - previous_token_time)
                            previous_token_time = now_time

                            # Stream off: Chunks are full response.
                            # chunks.append(chunk)

                        output = b"".join(chunks).decode("utf-8")
                        santicized = output.rstrip(
                            "\n\t "
                        )  # Remove trailing whitespace characters including EOF, and "[DONE]"
                    else:
                        time_to_first = time.perf_counter() - previous_token_time
                        output = await response.text()
                        santicized = output
                except Exception as e:
                    print_err(f"Failed to read response for request {idx}: {e}")
                    break
            try:
                ret = load_response(santicized)

                # Re-send the request if it failed.
                if "error" not in ret:
                    break
            except Exception as e:
                # It's ok to parse failure, santicized output could be jsonl, other format, or internal error.
                print_err(f"Invalid response for request {idx}: {santicized}: {e}")
                break

    request_end_time = time.perf_counter()
    request_latency = request_end_time - request_start_time

    if trace:
        request_trace = {
            "request_id": idx,
            "input_tokens": prompt_len,
            "output_tokens": output_len
            if len(token_latencies) == 0
            else len(token_latencies) + 1,
            "timestamp": ts.strftime("%Y-%m-%d %H:%M:%S %Z%z"),
            "E2E": request_latency,
            "status_code": status_code,
            "success": status_code == 200 if status_code else False,
        }
        if len(token_latencies) > 0:
            request_trace["TTFT"] = time_to_first
            request_trace["TPOT_mean"] = np.mean(token_latencies)  # type: ignore
            request_trace["TPOT_P50"] = np.percentile(token_latencies, 50)  # type: ignore
            request_trace["TPOT_P90"] = np.percentile(token_latencies, 90)  # type: ignore
            request_trace["TPOT_P99"] = np.percentile(token_latencies, 99)  # type: ignore
        print(json.dumps(request_trace))
    REQUEST_LATENCY.append((prompt_len, output_len, request_latency))
    if len(token_latencies) > 0:
        TOKEN_LATENCY.append((prompt_len, output_len, token_latencies))
    TIME_TO_FIRST_TOKEN.append(time_to_first)


async def benchmark(
    backend: str,
    api_url: str,
    api_key: Optional[str],
    model: str,
    input_requests: List[Tuple[str, int, int, float]],
    best_of: int,
    use_beam_search: bool,
    request_rate: float,
    num_requests: int,
    stream: bool,
    verbose: bool,
    trace: bool,
    use_workload_interval: bool = False,
) -> None:
    tasks: List[asyncio.Task] = []

    async for request in get_request(
        input_requests, request_rate, num_requests, verbose, use_workload_interval
    ):
        prompt, prompt_len, output_len, next_in = request
        task = asyncio.create_task(
            send_request(
                len(tasks),
                backend,
                api_url,
                api_key,
                model,
                prompt,
                prompt_len,
                output_len,
                next_in,
                best_of,
                use_beam_search,
                stream,
                verbose,
                trace,
            )
        )
        tasks.append(task)

    await asyncio.gather(*tasks)


def main(args: argparse.Namespace):
    # Set global temperature from args
    global TEMPERATURE
    TEMPERATURE = args.temperature

    result = {}
    if args.verbose:
        print(args)
    else:
        result["input_tokens"] = args.input_len
        result["output_tokens"] = args.output_len
        result["request_rate"] = args.request_rate
        result["seed"] = args.seed
        result["model"] = args.model
        result["temperature"] = args.temperature
        result["samples"] = args.num_prompts

    random.seed(args.seed)
    np.random.seed(args.seed)

    api_url = f"http://{args.host}:{args.port}/v1/completions"
    input_requests = sample_requests(
        args.num_prompts, args.input_len, args.output_len, args.workload_dataset_file
    )
    result["samples"] = len(input_requests)  # Update number of samples

    benchmark_start_time = time.perf_counter()
    try:
        asyncio.run(
            benchmark(
                args.backend,
                api_url,
                args.api_key,
                args.model,
                input_requests,
                args.best_of,
                args.use_beam_search,
                args.request_rate,
                len(input_requests),
                args.stream,
                args.verbose,
                args.trace,
                args.use_workload_interval,
            )
        )
    except Exception:
        import traceback

        traceback.print_exc()
    benchmark_end_time = time.perf_counter()
    benchmark_time = benchmark_end_time - benchmark_start_time

    if args.verbose:
        print()
        print("RESULT SUMMARY")
        print(f"Request rate: {args.request_rate} req/s")
        print(f"Prompt count: {len(REQUEST_LATENCY)}")
        print(f"Total time: {benchmark_time:.2f} s")
        print(
            f"Request Throughput: {len(REQUEST_LATENCY) / benchmark_time:.2f} requests/s"
        )
        print(
            f"Output Token Throughput: {sum([output for _, output, _ in REQUEST_LATENCY]) / benchmark_time:.2f} tokens/s"
        )
        print()
    elif not args.trace:
        result["metric"] = "TPUT"  # Throughput
        result["mean"] = len(REQUEST_LATENCY) / benchmark_time
        print(json.dumps(result))
        result["metric"] = "TT"  # Token throughput
        result["mean"] = (
            sum([output for _, output, _ in REQUEST_LATENCY]) / benchmark_time
        )
        print(json.dumps(result))

    # Compute the latency statistics.
    avg_latency = np.mean([latency for _, _, latency in REQUEST_LATENCY])
    if args.verbose:
        print("REQUEST LATENCIES")
        print(f"Avg: {avg_latency:.2f} s")
        print(
            f"50p: {np.percentile([latency for _, _, latency in REQUEST_LATENCY], 50)} s"
        )
        print(
            f"90p: {np.percentile([latency for _, _, latency in REQUEST_LATENCY], 90)} s"
        )
        print(
            f"99p: {np.percentile([latency for _, _, latency in REQUEST_LATENCY], 99)} s"
        )
        print()
    elif not args.trace:
        result["metric"] = "E2E"  # Request latency
        result["mean"] = avg_latency
        result["P50"] = np.percentile(
            [latency for _, _, latency in REQUEST_LATENCY], 50
        )
        result["P90"] = np.percentile(
            [latency for _, _, latency in REQUEST_LATENCY], 90
        )
        result["P99"] = np.percentile(
            [latency for _, _, latency in REQUEST_LATENCY], 99
        )
        print(json.dumps(result))

    if len(TOKEN_LATENCY) == 0:
        all_token_latencies = np.array([0.0])
    else:
        all_token_latencies = np.array(
            [
                latency
                for _, _, token_latencies in TOKEN_LATENCY
                for latency in token_latencies
            ]
        )
    if args.verbose:
        print("TOKEN LATENCIES")
        print("TTFT")
        print(f"Avg: {np.mean(TIME_TO_FIRST_TOKEN)}")
        print(f"50p: {np.percentile(TIME_TO_FIRST_TOKEN, 50)}")
        print(f"90p: {np.percentile(TIME_TO_FIRST_TOKEN, 90)}")
        print(f"99p: {np.percentile(TIME_TO_FIRST_TOKEN, 99)}")
        print("TPOT")
        print(f"Avg: {np.mean(all_token_latencies)}")
        print(f"50p: {np.percentile(all_token_latencies, 50)}")
        print(f"90p: {np.percentile(all_token_latencies, 90)}")
        print(f"99p: {np.percentile(all_token_latencies, 99)}")
        print()
    elif not args.trace:
        result["metric"] = "TTFT"  # Time to first token
        result["mean"] = np.mean(TIME_TO_FIRST_TOKEN)
        result["P50"] = np.percentile(TIME_TO_FIRST_TOKEN, 50)
        result["P90"] = np.percentile(TIME_TO_FIRST_TOKEN, 90)
        result["P99"] = np.percentile(TIME_TO_FIRST_TOKEN, 99)
        print(json.dumps(result))
        result["metric"] = "TPOT"  # Token latency
        result["mean"] = np.mean(all_token_latencies)
        result["P50"] = np.percentile(all_token_latencies, 50)
        result["P90"] = np.percentile(all_token_latencies, 90)
        result["P99"] = np.percentile(all_token_latencies, 99)
        print(json.dumps(result))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Benchmark the online serving throughput."
    )
    parser.add_argument("--backend", type=str, default="vllm", choices=["vllm"])
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=8000)
    parser.add_argument("--model", type=str, default="llama2-7b")
    parser.add_argument(
        "--best-of",
        type=int,
        default=1,
        help="Generates `best_of` sequences per prompt and " "returns the best one.",
    )
    parser.add_argument("--use-beam-search", action="store_true")
    parser.add_argument(
        "--num-prompts", type=int, default=0, help="Number of prompts to process."
    )
    parser.add_argument(
        "--request-rate",
        type=float,
        default=float("inf"),
        help="Number of requests per second. If this is inf, "
        "then all the requests are sent at time 0. "
        "Otherwise, we use Poisson process to synthesize "
        "the request arrival times.",
    )
    parser.add_argument("--seed", type=int, default=0)
    parser.add_argument(
        "--trust-remote-code",
        action="store_true",
        help="trust remote code from huggingface",
    )
    parser.add_argument("--input-len", type=int, default=0)
    parser.add_argument("--output-len", type=int, default=0)
    parser.add_argument("--api-key", type=str, default=None)
    parser.add_argument(
        "--verbose", action="store_true", help="Print human readable info to stdout"
    )
    parser.add_argument(
        "--trace",
        action="store_true",
        help="Print request trace to stdout instead of statistics",
    )
    parser.add_argument("--stream", action="store_true", help="Enable stream request.")
    parser.add_argument(
        "--workload_dataset_file",
        type=str,
        default=None,
        help="Path to a JSON file containing prompts",
    )
    parser.add_argument("--use-workload-interval", action="store_true")
    parser.add_argument(
        "--temperature",
        type=float,
        default=0.0,
        help="Temperature for text generation (default: 0.0)",
    )
    args = parser.parse_args()
    main(args)



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import threading
from datetime import datetime
from typing import Any, Callable, List, Optional, Tuple

import dash
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.graph_objs as go
from dash import dcc, html
from dash.dependencies import Input, Output
from starlette.middleware.wsgi import WSGIMiddleware
from starlette.routing import Mount

from .load_reader import DatasetLoadReader, LoadReader, WorkloadReader
from .monitor import DeploymentStates, ModelMonitor
from .profile_reader import FileProfileReader, ProfileReader, RedisProfileReader

canvas_size = 1000
scale = 16
interval = 1000  # in milliseconds
reader_interval = 10  # in seconds
interval_scaling = interval / 1000 / reader_interval

# window = 4000
# # Load dataset reader
# directory = os.path.dirname(os.path.abspath(__file__))
# reader: LoadReader = DatasetLoadReader(directory + '/data/sharegpt.csv', n_batch=100)
# span = window / reader.n_batch
# # Define clusterer
# clusterers: List[Clusterer] = [MovingDBSCANClusterer(0.8, 100, 4, window), DBSCANClusterer(0.5, 10)]
# Simulated data source for display
# data = DataBuffer(window)
# lvl2data = DataBuffer(window)

colors = [
    "red",
    "green",
    "pink",
    "blue",
    "navy",
    "orange",
    "purple",
    "cyan",
    "magenta",
    "yellow",
    "black",
    "gray",
    "brown",
    "olive",
    "teal",
    "maroon",
]


def default_datasource(_: str) -> Optional[ModelMonitor]:
    return get_debug_model_montior(None)


debug_monitor = None
figure = type("", (), {})()  # create a empty object
figure.__dict__ = {
    "debug_run": True,
    "debug_driver": None,
    "datasource": default_datasource,
    "last": dash.no_update,
    "lock": threading.Lock(),
}

logger = logging.getLogger("aibrix.gpu_optimizer.load_monitor.visualizer")


def get_debug_model_montior(
    path: Optional[str],
    is_workload: bool = True,
    scale: float = 1.0,
    profiles: Optional[List[str]] = None,
    redisprofile: Optional[str] = None,
) -> Optional[ModelMonitor]:
    global debug_monitor

    if debug_monitor is None:
        if path is None:
            directory = os.path.dirname(os.path.abspath(__file__))
            path = directory + "/data/sharegpt.csv"

        loadReader: Optional[LoadReader] = None
        if is_workload:
            loadReader = WorkloadReader(path, scale=scale, interval=reader_interval)
        else:
            loadReader = DatasetLoadReader(
                path, rps=10, scale=scale, interval=reader_interval
            )

        profile_reader: Optional[ProfileReader] = None
        if profiles is not None:
            profile_reader = FileProfileReader(profiles)
        elif redisprofile is not None:
            profile_reader = RedisProfileReader(
                *parse_redis_connection_str(redisprofile)
            )

        debug_monitor = ModelMonitor(
            "model", "0", loadReader, profile_reader=profile_reader, debug=True
        )
        if profile_reader is not None:
            for _, profile in enumerate(profile_reader.read()):
                debug_monitor.add_deployment(
                    "0", profile.gpu, None, DeploymentStates(profile.gpu, 0)
                )

    return debug_monitor


def parse_redis_connection_str(connection_str: str) -> Tuple[Any, str]:
    from urllib.parse import parse_qs, urlparse

    import redis

    # Parse the Redis URL
    url = urlparse(connection_str)

    # Connect to the Redis server
    db_name = str(url.path).strip("/")
    if db_name == "":
        db_name = "0"
    redis_client = redis.Redis(
        host=str(url.hostname),
        port=6379 if url.port is None else int(url.port),
        db=int(db_name),
        username=url.username,
        password=url.password,
    )

    # Store the result in Redis
    query_params = parse_qs(url.query)
    model_name = query_params.get("model", [""])[0]
    if model_name == "":
        raise Exception('"model" in Redic connection arguments is not provided.')

    return redis_client, model_name


def make_color(color, alpha=1):
    rgb = plt.matplotlib.colors.to_rgb(color)
    return f"rgba({rgb[0]*255}, {rgb[1]*255}, {rgb[2]*255}, {alpha})"


def update_graph(n, model_name):
    # Reset initial figure
    if n == 0:
        figure.last = dash.no_update

    # Acquire the lock at the beginning of the callback
    if not figure.lock.acquire(blocking=False):
        # If the lock is already acquired, skip this execution
        return figure.last

    try:
        start = datetime.now().timestamp()

        monitor: Optional[ModelMonitor] = figure.datasource(model_name)
        if monitor is None:
            figure.last = {
                "data": [],
                "layout": go.Layout(
                    title=f"Live data update of {model_name} is unavailable: model not monitored",
                    xaxis=dict(range=[0, scale], title="output_tokens(log2)"),
                    yaxis=dict(range=[0, scale], title="input_tokens(log2)"),
                ),
            }
            return figure.last

        if figure.debug_run:
            # Drive the monitor progress for debugging
            if figure.debug_driver is None:
                figure.debug_driver = monitor._run_yieldable(
                    True, window_scaling=interval_scaling
                )

            next(figure.debug_driver)

        data_df = monitor.dataframe
        if data_df is None or len(data_df) == 0:
            figure.last = {
                "data": [],
                "layout": go.Layout(
                    title=f"Live data update of {model_name} is unavailable: insufficient data",
                    xaxis=dict(range=[0, scale], title="output_tokens(log2)"),
                    yaxis=dict(range=[0, scale], title="input_tokens(log2)"),
                ),
            }
            return figure.last
        centers = monitor.centers
        labeled = monitor.labeled
        data_colors = [
            colors[int(label) % len(colors)] if label >= 0 else "black"
            for label in data_df["label"]
        ]
        # label_seen = len(centers)

        # recluster level 2
        # lvl2data.clear()
        # lvl2data.append(uncategorized)
        # clusterers[1].reset()
        # clusterers[1].insert(uncategorized)
        # lvl2labels, lvl2centers = clusterers[1].get_cluster_labels(lvl2data.xy)
        # labeled = reduce(lambda cnt, center: cnt+center.size, lvl2centers, labeled)
        # for i, label in enumerate(lvl2labels):
        #     if label < 0:
        #         continue
        #     lvl2data._color[i] = colors[(int(label)+label_seen) % len(colors)]
        # label_seen += len(lvl2centers)
        # if len(lvl2centers) > 0:
        #     center_df = pd.concat([center_df, pd.DataFrame(data=np.array([center.to_array(span) for center in lvl2centers]), columns=['x', 'y', 'radius', 'size'])], ignore_index=True)

        duration = (datetime.now().timestamp() - start) * 1000
        plotdata = [
            go.Scatter(
                x=data_df["input_tokens"],
                y=data_df["output_tokens"],
                mode="markers",
                name="major patterns",
                marker=dict(
                    color=data_colors,  # Specify the color of the marker
                    size=3,  # Set the size of the marker
                ),
            ),
            # go.Scatter(
            #     x=lvl2data.x,
            #     y=lvl2data.y,
            #     mode='markers',
            #     name='minor patterns',
            #     marker=dict(
            #         symbol='square',
            #         color=lvl2data.color,  # Specify the color of the marker
            #         size=3,            # Set the size of the marker
            #     )
            # ),
        ]
        if len(centers) > 0:
            center_df = pd.DataFrame(
                data=np.array([center.to_array() for center in centers]),
                columns=["y", "x", "radius", "size"],
            )
            # assign color to center_df
            center_colors = [
                make_color(colors[int(idx) % len(colors)], alpha=0.5)
                for idx in center_df.index
            ]
            # print(center_df['size'])
            plotdata.append(
                go.Scatter(
                    x=center_df["x"],
                    y=center_df["y"],
                    mode="markers",
                    name="RPS",
                    marker=dict(
                        sizeref=1,  # Adjust this value to control size
                        sizemode="diameter",
                        size=np.maximum(
                            (canvas_size / (scale + 2))
                            * (np.log2(center_df["size"]) + 1),
                            10,
                        ),  # Assuming you have a column with size values
                        color=center_colors,
                        symbol="circle",
                    ),
                )
            )

        figure.last = {
            "data": plotdata,
            "layout": go.Layout(
                title=f"Live Data Update({n}:{round(duration)}ms) of {model_name}, labeled: {round(labeled/len(data_df)*100, 2)}%, processed: {monitor.progress}",
                # xaxis=dict(range=[0, max(data['x']) + 1]),
                # yaxis=dict(range=[0, max(data['y']) + 1])
                xaxis=dict(range=[0, scale], title="input_tokens(log2)"),
                yaxis=dict(range=[0, scale], title="output_tokens(log2)"),
            ),
        }
        return figure.last
    except Exception as e:
        logger.error(f"Failed to prepare figure: {e}")
        import traceback

        traceback.print_exc()
    finally:
        # Release the lock at the end of the callback
        figure.lock.release()


def store_model_name(pathname):
    # Extract model_name from pathname (e.g., /dash/model_name/)
    try:
        model_name = pathname.strip("/").split("/")[-1]
    except IndexError:
        model_name = None  # Handle cases where model_name is not present
    return model_name


def init(prefix=""):
    app = dash.Dash(__name__, requests_pathname_prefix=prefix + "/")

    app.layout = html.Div(
        [
            dcc.Location(id="url", refresh=False),  # To access the URL
            dcc.Input(id="model-name-input", type="hidden", value=""),
            html.Div(id="model-info"),
            dcc.Interval(
                id="interval-component",
                interval=1000,  # in milliseconds
                n_intervals=0,  # start at 0
            ),
            dcc.Graph(
                id="live-graph",
                style={"width": f"{canvas_size}px", "height": f"{canvas_size}px"},
            ),
        ]
    )

    app.callback(Output("model-name-input", "value"), Input("url", "pathname"))(
        store_model_name
    )
    app.callback(
        Output("live-graph", "figure"),
        [
            Input("interval-component", "n_intervals"),
            Input("model-name-input", "value"),
        ],
    )(update_graph)

    return app


def mount_to(
    routes: List, prefix: str, datasrc: Callable[[str], Optional[ModelMonitor]]
):
    figure.datasource = datasrc
    figure.debug_run = False
    routes.append(Mount(prefix, WSGIMiddleware(init(prefix).server)))
    return routes


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger("pulp.apis.core").setLevel(logging.INFO)  # Suppress pulp logs

    import argparse

    parser = argparse.ArgumentParser(
        description="Please provide dataset or workload path:"
    )
    parser.add_argument(
        "--dataset",
        type=str,
        default=None,
        help="Dataset path. A Dataset has no timestamp info.",
    )
    parser.add_argument("--workload", type=str, default=None, help="Workload path.")
    parser.add_argument(
        "--scaledata",
        type=float,
        default=1,
        help="A factor to scale the number of input/output tokens.",
    )
    parser.add_argument("--profile", type=str, default=None, help="Profile path.")
    parser.add_argument(
        "--profiles", nargs="+", type=str, default=None, help="Profile path."
    )
    parser.add_argument(
        "--redisprofile",
        type=str,
        default=None,
        help="Redis connection string for profiles.",
    )
    args = parser.parse_args()
    filepath, is_workload = (
        (args.workload, True) if args.workload is not None else (args.dataset, False)
    )
    if filepath is not None:
        figure.datasource = lambda _: get_debug_model_montior(
            filepath,
            is_workload,
            args.scaledata,
            profiles=args.profiles
            if args.profiles is not None
            else [args.profile]
            if args.profile is not None
            else None,
            redisprofile=args.redisprofile,
        )
    init().run_server(debug=True)



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from aibrix.downloader.base import get_downloader


def download_model(
    model_uri: str,
    local_path: Optional[str] = None,
    model_name: Optional[str] = None,
    download_extra_config: Optional[Dict] = None,
    enable_progress_bar: bool = False,
):
    """Download model from model_uri to local_path.

    Args:
        model_uri (str): model uri.
        local_path (str): local path to save model.
    """

    downloader = get_downloader(
        model_uri, model_name, download_extra_config, enable_progress_bar
    )
    return downloader.download_model(local_path)


__all__ = ["download_model", "get_downloader"]



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from typing_extensions import TypeAlias

DataSignatures: TypeAlias = np.ndarray
"""ndarray of shape(n, 2)"""

DataSignature: TypeAlias = np.ndarray
"""ndarray of shape(2,)"""


class DataPoint(np.ndarray):
    def __new__(
        cls,
        *args,
        age: Union[int, float] = 0,
        ndarray: Optional[np.ndarray] = None,
        **kwargs,
    ):
        if ndarray is not None:
            return ndarray.view(cls)

        ins = np.empty((3,))
        if len(args) > 0:
            ins[0] = args[0]
        if len(args) > 1:
            ins[1] = args[1]
        ins[2] = age
        return ins

    @property
    def age(self):
        return self[2]

    @property
    def signature(self) -> DataSignature:
        return self[:2]


class DataPoints(np.ndarray):
    def __new__(cls, ndarray: np.ndarray):
        return ndarray.view(cls)

    @property
    def signatures(self) -> DataSignatures:
        return self[:, :2]

    def datapoint(self, idx):
        return DataPoint(ndarray=self[idx])


class DataBuffer:
    def __init__(self, cap: int):
        self._xy = np.empty((cap, 3), dtype=float)
        self._commited = 0
        """The length of data that has been processed and ready to read"""
        self._head = 0
        """All length of all data that includes processing data points."""

    def reconcile(self, cap: int):
        if cap < self._xy.shape[0]:
            # We do not shrink
            return

        new_cap = self._xy.shape[0] * 2
        while new_cap < cap:
            new_cap *= 2
        self._xy = np.resize(self._xy, (new_cap, 3))

    def append(self, tokens: List[DataPoint], commit: bool = False) -> DataPoints:
        """Append data points to the buffer. If commit is True, the data points will be committed immediately and could lead to data inconsistent if it takes a long time to process new data."""
        size_gap = self._commited + len(tokens) - self._xy.shape[0]
        # Check buffer size.
        if size_gap > 0:
            # We do not expand the buffer automatically, simply evicting some records to make room for new data.
            self.trim_head(size_gap)

        # Check tokens size. If tokens is too large, buffer has been cleared at this point.
        if len(tokens) > self._xy.shape[0]:
            tokens = tokens[-self._xy.shape[0] :]

        self._xy[self._commited : self._commited + len(tokens)] = tokens
        ret = DataPoints(self._xy[self._commited : self._commited + len(tokens)])
        self._head += len(tokens)
        if commit:
            self.commit()
        return ret

    def commit(self):
        self._commited = self._head

    def trim_head(self, start):
        if self._head != self._commited:
            raise Exception("Cannot trim head when there are uncommited data points.")

        if start >= self._commited:
            # empty(), skip data moving
            self._head = self._commited = 0
            return
        elif start <= -self._commited:
            # keep all data
            return

        # Do compacting.
        # implement self._xy[:-start] = self._xy[start:] with variable buffer length.
        if start > 0:
            self._xy[: self._commited - start] = self._xy[start : self._commited]
            self._commited -= start
        else:
            self._xy[:-start] = self._xy[self._commited + start : self._commited]
            self._commited = -start
        self._head = self._commited

    def clear(self):
        self._commited = 0
        self._head = 0

    @property
    def x(self):
        return self._xy[: self._commited, 0]

    @property
    def y(self):
        return self._xy[: self._commited, 1]

    @property
    def datapoints(self) -> DataPoints:
        return DataPoints(self._xy[: self._commited])

    @property
    def len(self):
        return self._commited

    @property
    def cap(self):
        return self._xy.shape[0]


class Centeroid:
    def __init__(self, span=0):
        """Centeroid calculates the mass center, radius, and size of data points."""
        self._sum_center = None
        self._range_max = None
        self._range_min = None
        self._span_max = 0
        self._span_min = 0
        self._span = span  # Used to override span if set
        self._size = 0
        self._signature = None
        self._up_to_date = False  # Whether the signature is up to date.
        self._signature_tolerance = 0.5
        """Assuming indexes are of equal distances, the signature_tolerance 
        control how a value in the middle of two indexes should aligned to a 
        certain index. While 0.5 indicate neutual preference. 0.25 set 
        threshold closer to lower index and then prefer higher index."""

    def add(self, point: DataPoint):
        if self._sum_center is None:
            self._sum_center = list(point.signature)
            self._range_min = list(point.signature)
            self._range_max = list(point.signature)
            self._span_max = point.age
            self._span_min = point.age
            self._signature = np.zeros_like(self._sum_center, dtype=int)
        else:
            for i, val in enumerate(point.signature):  # type: ignore
                self._sum_center[i] += val
                self._range_min[i] = min(self._range_min[i], val)
                self._range_max[i] = max(self._range_max[i], val)
            self._span_min = min(self._span_min, point.age)
            self._span_max = max(self._span_max, point.age)
            self._up_to_date = False

        self._size += 1

    def get_signature(
        self,
        indexes: List[List[float]],
        error_suppressor: Optional[
            Callable[[int, float, float, float, float], bool]
        ] = None,
    ) -> Tuple[int]:
        """Generate the index signature of the centroid within the indexes' range.

        Args:
            indexes: A list of list of float, each list is a range of values.
            error_suppressor: A function to handle the error with parameters(value, index assigned, value of index, offset). Return True to accept closest value. If None, raise an exception.
        """
        if len(self._signature) != len(indexes):
            raise Exception(
                f"Indexes and centeroid signature size mismatch, {len(self._signature)}:{len(indexes)}"
            )

        if self._up_to_date:
            return self.signature

        for i, value in enumerate(self.center):
            if len(indexes[i]) == 0:
                raise Exception("Indexes size mismatch, at least 1.")
            elif len(indexes[i]) == 1:
                self._signature[i] = 0
                continue

            # Assuming indexes are ascending ordered.
            distance = (indexes[i][-1] - indexes[i][0]) / (len(indexes[i]) - 1)
            if value < indexes[i][0]:
                offset = -distance * (1 - self._signature_tolerance)
                if value >= indexes[i][0] + offset:
                    # with in threshold, no warning
                    self._signature[i] = 0
                elif error_suppressor is not None:
                    # error suppressor is available, let error_suppressor decide how to proceed.
                    if error_suppressor(i, value, 0, indexes[i][0], offset):
                        self._signature[i] = 0
                else:
                    raise Exception(
                        f"Centeroid is out of range: {i}:{value} and accepted minimum {indexes[i][0]} (with offset {offset})."
                    )
            elif value > indexes[i][-1]:
                offset = distance * self._signature_tolerance
                if value <= indexes[i][0] + offset:
                    # with in threshold, no warning
                    self._signature[i] = len(indexes[i]) - 1
                elif error_suppressor is not None:
                    # error suppressor is available, let error_suppressor decide how to proceed.
                    if error_suppressor(
                        i, value, len(indexes[i]) - 1, indexes[i][-1], distance
                    ):
                        self._signature[i] = len(indexes[i]) - 1
                else:
                    raise Exception(
                        f"Centeroid is out of range: {i}:{value} and accepted maximum {indexes[i][-1]} (with offset {offset})."
                    )
            else:
                # Find the index using binary search.
                left, right = 0, len(indexes[i]) - 1
                found = False
                while left < right - 1:
                    mid = (left + right) // 2
                    if value < indexes[i][mid]:
                        right = mid
                    elif value > indexes[i][mid]:
                        left = mid
                    else:
                        self._signature[i] = mid
                        found = True
                        break
                if not found:
                    self._signature[i] = (
                        left
                        if value
                        < indexes[i][left]
                        + (indexes[i][right] - indexes[i][left])
                        * self._signature_tolerance
                        else right
                    )

        self._up_to_date = True
        return self.signature

    @property
    def center(self):
        return tuple(val / self._size for val in self._sum_center)

    @property
    def radius(self):
        return max(
            (val - self._range_min[i]) / 2 for i, val in enumerate(self._range_max)
        )

    @property
    def size(self):
        return self._size

    @property
    def span(self):
        return (
            self._span
            if self._span > 0
            else 1.0
            if self._span_max == self._span_min
            else self._span_max - self._span_min
        )

    @span.setter
    def span(self, value):
        self._span = value

    @property
    def signature(self) -> Tuple[int]:
        if not self._up_to_date:
            raise Exception("Signature is not up to date.")
        return tuple(self._signature.tolist())

    @property
    def rate(self):
        return self._size / self.span

    def to_array(self):
        ret = list(self.center)
        ret.append(self.radius)
        ret.append(self.rate)
        return ret

    def __str__(self) -> str:
        return f"Centeroid[center={self.center}, rps={self.rate}]"



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from functools import reduce
from typing import Iterable, Optional, Tuple

import numpy as np

from aibrix.gpu_optimizer.utils import DelayedLog

from .solver.melange import Config as MelangConfig
from .solver.melange import SolverRunner
from .types import GPUProfile, WorkloadProfile

logger = logging.getLogger("aibrix.gpuoptimizer.optimizer")


class Optimizer:
    def __init__(
        self, gpu_fraction: float, profiles: Optional[Iterable[GPUProfile]] = None
    ):
        self._config = MelangConfig()
        self._workload_distribution_template: Optional[np.ndarray] = None
        self._indexes: Optional[list] = None  # Values ticks of tputs columns and rows
        self._log_indexes: Optional[list] = None  # Cache the log2 value of index
        self._gpu_fraction = gpu_fraction
        if profiles is not None:
            for profile in profiles:
                self.set_profile(profile)

    def set_profile(self, profile: GPUProfile):
        if self._workload_distribution_template is None:
            self._workload_distribution_template = np.zeros_like(profile.tputs)
            self._indexes = profile.indexes
            if profile.indexes is not None:
                self._log_indexes = [
                    np.log2(index).tolist() for index in profile.indexes
                ]
        elif (
            self._workload_distribution_template.shape != np.shape(profile.tputs)
            or self._indexes != profile.indexes
        ):
            raise Exception(
                f"Profile({profile.gpu}) applied should keep a same shape and value ticks. shapes: {self._workload_distribution_template.shape} vs {np.shape(profile.tputs)}, indexes: {self._indexes} vs {profile.indexes}"
            )

        logger.debug(
            "Applied profile for %s, shape: %s, indees: %s",
            profile.gpu,
            profile.tputs,
            self._indexes,
        )
        self._config.gpu_info[profile.gpu] = profile.__dict__

    def delete_profile(self, gpu):
        if gpu in self._config.gpu_info:
            del self._config.gpu_info[gpu]

    def set_workload_distribution(
        self, profiles: Iterable[WorkloadProfile], total_request_rate: float
    ) -> bool:
        """Update workload distribution and return success or failure."""
        if self._workload_distribution_template is None:
            return False
        self._workload_distribution_template.fill(0)

        # Maintain the overall request scale disregard some request are not covered.
        self._config.total_request_rate = total_request_rate * self._gpu_fraction
        # covered_request_rate is used to calculate the workload distribution.
        covered_request_rate = reduce(
            lambda cnt, center: cnt + center.rate, profiles, 0.0
        )
        success = True
        for profile in profiles:
            try:
                signature = self._validate_workload_signature(profile)
                # Merge possible multiple patterns (out of range patterns coinincident with border patterns)
                self._workload_distribution_template[signature] += (
                    profile.rate / covered_request_rate
                )  # type: ignore
                logger.debug(
                    f"Resolved {profile} to signature={signature}, output-input={DelayedLog(lambda: self._log_signature_expr(signature))}, modified-rps={self._workload_distribution_template[signature]*total_request_rate}, overall-rps={total_request_rate}, capacity={DelayedLog(lambda: self._log_capacity(signature))}"
                )
            except Exception as e:
                logger.error(
                    f"Fail to set workload distribution: {profile.signature}: {e}"
                )
                success = False
        self._config.workload_distribution = (
            self._workload_distribution_template.tolist()
        )
        return success

    def run(self) -> Optional[dict]:
        """Run the solver and return the result.
        Return None if no profiles are added.
        The result is a dict with the following format:

        {
            "gpu1": replicas1,
            "gpu1": replicas2,
            "cost": cost,
        }
        """
        logger.debug(f"Starting solver for {self._config.gpu_info.keys()}")
        if len(self._config.gpu_info) == 0:
            return None

        runner = SolverRunner(self._config)
        ret = runner.run()
        logger.debug(f"Done solver: {ret}")
        return ret

    def _validate_workload_signature(self, profile: WorkloadProfile) -> Tuple[int]:
        """Validate workload's signature by regard each element in signature tuple a index.
        return valid index tuple for accessing  self._workload_distribution_template"""
        if (
            self._workload_distribution_template is None
            or self._indexes is None
            or self._log_indexes is None
        ):
            raise Exception("Load profile not set.")

        signature = profile.get_signature(self._log_indexes, self._log_signature_error)
        if len(signature) != self._workload_distribution_template.ndim:
            raise Exception(
                f"Unmatch workload profile, expected a signature of length {self._workload_distribution_template.ndim} , got {len(signature)}."
            )

        # No validation on the shape. Leave set function to throw error
        return signature

    def _log_signature_error(
        self, dimeansion, value, index, index_value, offset
    ) -> bool:
        logger.warning(
            f"Signature item {dimeansion}:{value} is out of range, counted as{index_value} (reference offset: {offset})"
        )
        return True

    def _log_signature_expr(self, signature: Tuple[int]) -> str:
        if self._indexes is None:
            return "_index not set"

        values = list(signature)
        for i, value in enumerate(values):
            values[i] = self._indexes[i][value]
        return str(tuple(values))

    def _log_capacity(self, signature: Tuple[int]) -> str:
        return str(
            tuple(
                (
                    np.array(gpu_info["tputs"])[signature]
                    for gpu_info in self._config.gpu_info.values()
                )
            )
        )



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from aibrix.batch.storage.generic_storage import LocalDiskFiles, StorageType

# [TODO] Add S3 as another storage
from aibrix.batch.storage.tos_storage import TOSStorage

current_job_offsets = {}
job_input_requests = {}
p_storage = None
NUM_REQUESTS_PER_READ = 1024


def initialize_batch_storage(storage_type=StorageType.LocalDiskFile, params={}):
    """Initialize storage type. Now it support files and TOS.

    For some storage type, user needs to pass in other parameters to params.
    """
    global p_storage
    if storage_type == StorageType.LocalDiskFile:
        p_storage = LocalDiskFiles()
    elif storage_type == StorageType.TOS:
        p_storage = TOSStorage(**params)
    else:
        raise ValueError("Unknown storage type")


def upload_input_data(inputDataFileName):
    """Upload job input data file to storage.

    Args:
        inputDataFileName (str): an input file string.
    """
    job_id = uuid.uuid1()
    p_storage.write_job_input_data(job_id, inputDataFileName)

    current_job_offsets[job_id] = 0
    return job_id


def read_job_requests(job_id, start_index, num_requests):
    """Read job requests starting at index: start_index.

    Instead of reading from storage per request, this maintains a list of requests
    in memory with a length of NUM_REQUESTS_PER_READ.
    It also supports random access, if backward read is necessary.
    """

    if job_id not in current_job_offsets:
        print(f"Create job {job_id} first. Can not find corresponding job ID!!")
        return []

    # if no request is cached, this reads a list of requests from storage.
    if job_id not in job_input_requests:
        request_inputs = p_storage.read_job_input_data(job_id, 0, NUM_REQUESTS_PER_READ)
        job_input_requests[job_id] = request_inputs

    current_start_idx = current_job_offsets[job_id]
    current_end_idx = current_start_idx + len(job_input_requests[job_id])

    # this reads request backward, so it only reads necessary part.
    if start_index < current_start_idx:
        if start_index + num_requests < current_start_idx + 1:
            temp_len = num_requests
            diff_requests = p_storage.read_job_input_data(job_id, start_index, temp_len)
            job_input_requests[job_id] = diff_requests
        else:
            temp_len = current_start_idx + 1 - start_index
            diff_requests = p_storage.read_job_input_data(job_id, start_index, temp_len)
            job_input_requests[job_id] = diff_requests + job_input_requests[job_id]

        current_job_offsets[job_id], current_start_idx = start_index, start_index
        current_end_idx = current_start_idx + len(job_input_requests[job_id])

    # the cached parts miss already, this throws away old caches.
    if start_index >= current_end_idx:
        current_job_offsets[job_id] = start_index
        job_input_requests[job_id] = []
        current_start_idx, current_end_idx = start_index, start_index

    # now this reads necessary requests at least for a length of num_requests.
    if start_index + num_requests > current_end_idx:
        temp_len = start_index + num_requests - current_end_idx
        diff_requests = p_storage.read_job_input_data(job_id, current_end_idx, temp_len)
        job_input_requests[job_id] = job_input_requests[job_id] + diff_requests
        current_end_idx = current_job_offsets[job_id] + len(job_input_requests[job_id])

    available_num_req = min(num_requests, current_end_idx - start_index)
    start_offset = start_index - current_start_idx

    requests = job_input_requests[job_id][
        start_offset : start_offset + available_num_req
    ]
    # print("debug", len(requests))
    return requests


def put_storage_job_results(job_id, start_index, requests_results):
    """Write job results on a specific index."""
    p_storage.write_job_output_data(job_id, start_index, requests_results)


def get_storage_job_results(job_id, start_index, num_requests):
    """Read job requests results."""
    return p_storage.read_job_output_data(job_id, start_index, num_requests)


def remove_storage_job_data(job_id):
    """Remove job all relevant data."""
    p_storage.delete_job_data(job_id)


def get_job_request_len(job_id):
    """Get the number of requests for the job_id."""
    return p_storage.get_job_number_requests(job_id)



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

from aibrix.metrics.standard_rules import RenameStandardRule, StandardRule

# Standard rule accroding to https://docs.google.com/document/d/1SpSp1E6moa4HSrJnS4x3NpLuj88sMXr2tbofKlzTZpk
VLLM_METRIC_STANDARD_RULES: Dict[str, StandardRule] = {
    "vllm:request_success": RenameStandardRule(
        "vllm:request_success", "aibrix:request_success"
    ),
    "vllm:num_requests_waiting": RenameStandardRule(
        "vllm:num_requests_waiting", "aibrix:queue_size"
    ),
    "vllm:time_to_first_token_seconds": RenameStandardRule(
        "vllm:time_to_first_token_seconds", "aibrix:time_to_first_token_seconds"
    ),
    "vllm:gpu_cache_usage_perc": RenameStandardRule(
        "vllm:gpu_cache_usage_perc", "aibrix:gpu_cache_usage_perc"
    ),
    "vllm:time_per_output_token_seconds": RenameStandardRule(
        "vllm:time_per_output_token_seconds", "aibrix:time_per_output_token"
    ),
    "vllm:e2e_request_latency_seconds": RenameStandardRule(
        "vllm:e2e_request_latency_seconds", "aibrix:e2e_request_latency"
    ),
}

# TODO add more engine standard rules


def get_metric_standard_rules(engine: str) -> Dict[str, StandardRule]:
    if engine.lower() == "vllm":
        return VLLM_METRIC_STANDARD_RULES
    else:
        raise ValueError(f"Engine {engine} is not supported.")



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import requests
from prometheus_client.parser import text_string_to_metric_families
from prometheus_client.registry import Collector

from aibrix.config import DEFAULT_METRIC_COLLECTOR_TIMEOUT
from aibrix.logger import init_logger
from aibrix.metrics.standard_rules import StandardRule

logger = init_logger(__name__)


class HTTPCollector(Collector):
    def __init__(
        self,
        endpoint: str,
        metrics_rules: Dict[str, StandardRule],
        keep_original_metric: bool = True,
        timeout=DEFAULT_METRIC_COLLECTOR_TIMEOUT,
    ):
        self.metric_endpoint = endpoint
        self.metrics_rules = metrics_rules
        self.keep_original_metric = keep_original_metric

        self.timeout = timeout
        self.session = requests.Session()

    def _collect(self):
        try:
            response = self.session.get(self.metric_endpoint, timeout=self.timeout)
            if response.status_code != 200:
                logger.warning(
                    f"Failed to collect metrics from {self.metric_endpoint} "
                    f"with status code {response.status_code}, "
                    f"response: {response.text}"
                )
                return ""
            return response.text
        except Exception as e:
            logger.warning(
                f"Failed to collect metrics from {self.metric_endpoint}: {e}"
            )
            return ""

    def collect(self):
        metrics_text = self._collect()
        for m in text_string_to_metric_families(metrics_text):
            if self.keep_original_metric:
                yield m

            # metric standardizing rule matched
            if m.name in self.metrics_rules:
                new_metric = self.metrics_rules[m.name](m)
                if new_metric is not None:
                    yield from new_metric



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import contextlib
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Generator, List, Optional

from filelock import BaseFileLock, FileLock, Timeout

from aibrix.config import DOWNLOAD_CACHE_DIR, DOWNLOAD_FILE_LOCK_CHECK_TIMEOUT
from aibrix.logger import init_logger

logger = init_logger(__name__)


class RemoteSource(Enum):
    S3 = "s3"
    TOS = "tos"
    HUGGINGFACE = "huggingface"
    UNKNOWN = "unknown"

    def __str__(self):
        return self.value


class FileDownloadStatus(Enum):
    DOWNLOADING = "downloading"
    DOWNLOADED = "downloaded"
    NO_OPERATION = "no_operation"  # Interrupted from downloading
    UNKNOWN = "unknown"

    def __str__(self):
        return self.value


class ModelDownloadStatus(Enum):
    NOT_EXIST = "not_exist"
    DOWNLOADING = "downloading"
    DOWNLOADED = "downloaded"
    NO_OPERATION = "no_operation"  # Interrupted from downloading
    UNKNOWN = "unknown"

    def __str__(self):
        return self.value


@dataclass
class DownloadFile:
    file_path: Path
    lock_path: Path
    metadata_path: Path

    @property
    def status(self):
        if self.file_path.exists() and self.metadata_path.exists():
            return FileDownloadStatus.DOWNLOADED

        try:
            # Downloading process will acquire the lock,
            # and other process will raise Timeout Exception
            lock = FileLock(self.lock_path)
            lock.acquire(blocking=False)
        except Timeout:
            return FileDownloadStatus.DOWNLOADING
        except Exception as e:
            logger.warning(f"Failed to acquire lock failed for error: {e}")
            return FileDownloadStatus.UNKNOWN
        else:
            return FileDownloadStatus.NO_OPERATION

    @contextlib.contextmanager
    def download_lock(self) -> Generator[BaseFileLock, None, None]:
        """A filelock that download process should be acquired.
        Same implementation as WeakFileLock in huggingface_hub."""
        lock = FileLock(self.lock_path, timeout=DOWNLOAD_FILE_LOCK_CHECK_TIMEOUT)
        while True:
            try:
                lock.acquire()
            except Timeout:
                logger.info(
                    f"Still waiting to acquire download lock on {self.lock_path}"
                )
            else:
                break

        yield lock

        try:
            return lock.release()
        except OSError:
            try:
                Path(self.lock_path).unlink()
            except OSError:
                pass


@dataclass
class DownloadModel:
    model_source: RemoteSource
    local_path: Path
    model_name: str
    download_files: List[DownloadFile]

    def __post_init__(self):
        if len(self.download_files) == 0:
            logger.warning(f"No download files found for model {self.model_name}")

    @property
    def status(self):
        all_status = [file.status for file in self.download_files]
        if all(status == FileDownloadStatus.DOWNLOADED for status in all_status):
            return ModelDownloadStatus.DOWNLOADED

        if any(status == FileDownloadStatus.DOWNLOADING for status in all_status):
            return ModelDownloadStatus.DOWNLOADING

        if all(
            status in [FileDownloadStatus.DOWNLOADED, FileDownloadStatus.NO_OPERATION]
            for status in all_status
        ):
            return ModelDownloadStatus.NO_OPERATION

        return ModelDownloadStatus.UNKNOWN

    @property
    def model_root_path(self) -> Path:
        return Path(self.local_path).joinpath(self.model_name)

    @classmethod
    def infer_from_model_path(
        cls, local_path: Path, model_name: str, source: RemoteSource
    ) -> Optional["DownloadModel"]:
        assert source is not None

        model_base_dir = Path(local_path).joinpath(model_name)

        # model not exists
        if not model_base_dir.exists():
            return None

        cache_sub_dir = (DOWNLOAD_CACHE_DIR % source.value).strip("/")
        cache_dir = Path(model_base_dir).joinpath(cache_sub_dir)
        lock_files = list(Path(cache_dir).glob("*.lock"))

        download_files = []
        for lock_file in lock_files:
            lock_name = lock_file.name
            lock_suffix = ".lock"
            if lock_name.endswith(lock_suffix):
                filename = lock_name[: -len(lock_suffix)]
                download_file = get_local_download_paths(
                    model_base_dir=model_base_dir, filename=filename, source=source
                )
                download_files.append(download_file)

        return cls(
            model_source=source,
            local_path=local_path,
            model_name=model_name,
            download_files=download_files,
        )

    @classmethod
    def infer_from_local_path(cls, local_path: Path) -> List["DownloadModel"]:
        models: List["DownloadModel"] = []

        def remove_suffix(input_string, suffix):
            if not input_string.endswith(suffix):
                return input_string
            return input_string[: -len(suffix)]

        for source in RemoteSource:
            cache_sub_dir = (DOWNLOAD_CACHE_DIR % source.value).strip("/")
            cache_dirs = list(Path(local_path).glob(f"**/{cache_sub_dir}"))
            if not cache_dirs:
                continue

            for cache_dir in cache_dirs:
                relative_path = cache_dir.relative_to(local_path)
                relative_str = str(relative_path).strip("/")
                model_name = remove_suffix(relative_str, cache_sub_dir).strip("/")

                download_model = cls.infer_from_model_path(
                    local_path, model_name, source
                )
                if download_model is None:
                    continue
                models.append(download_model)

        return models

    def __str__(self):
        return (
            "DownloadModel(\n"
            + f"\tmodel_source={self.model_source.value},\n"
            + f"\tmodel_name={self.model_name},\n"
            + f"\tstatus={self.status.value},\n"
            + f"\tlocal_path={self.local_path},\n"
            + f"\tdownload_files_count={len(self.download_files)}\n)"
        )


def get_local_download_paths(
    model_base_dir: Path, filename: str, source: RemoteSource
) -> DownloadFile:
    file_path = model_base_dir.joinpath(filename)
    sub_cache_dir = (DOWNLOAD_CACHE_DIR % source.value).strip("/")
    cache_dir = model_base_dir.joinpath(sub_cache_dir)
    lock_path = cache_dir.joinpath(f"{filename}.lock")
    metadata_path = cache_dir.joinpath(f"{filename}.metadata")
    return DownloadFile(file_path, lock_path, metadata_path)



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Callable
from urllib.parse import quote


class DelayedLog:
    def __init__(self, expr: Callable[[], Any]):
        self._expr = expr

    def __str__(self) -> str:
        return str(self._expr())

    def __expr__(self) -> str:
        return self.__str__()


class ExcludePathsFilter(logging.Filter):
    def __init__(self, exclude_paths):
        super().__init__()
        self.exclude_paths = [quote(exclude_path) for exclude_path in exclude_paths]

    def filter(self, record):
        # Check if the record is an access log and extract the path
        if hasattr(record, "args") and len(record.args) >= 3:
            request_path = record.args[2]
            return not any(request_path.startswith(path) for path in self.exclude_paths)
        return True  # Allow other log records



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import bisect
import queue
import time
from abc import ABC, abstractmethod
from enum import Enum

from aibrix.batch.constant import DEFAULT_JOB_POOL_SIZE, EXPIRE_INTERVAL
from aibrix.batch.job_manager import JobStatus


class SchedulePolicy(Enum):
    FIFO = 1


class CCInterface(ABC):
    @abstractmethod
    def update_job_pool_size(self, new_pool_size):
        pass

    @abstractmethod
    def shrink_resource(self):
        pass

    @abstractmethod
    def grow_resource(self):
        pass


class BasicCongestionControl(CCInterface):
    def __init__(self, pool_size):
        self._running_job_pool = [None] * pool_size
        self._job_pool_size = pool_size
        self._running_job_idx = 0

    def tighten_jobs(self):
        """
        We are intentional to move all current in-progress jobs
        to the front part of the pool.
        Two advantages of doing this:
        1. it facilitates the following resource adjustment by update the tail of the pool.
        2. it guarantees the jobs' order if necessary.
        """
        current_job_id = self._running_job_pool[self._running_job_idx]

        # Move all jobs to the front part of pool to maitain the order.
        slow_id, fast_id = 0, 0
        while slow_id < len(self._running_job_pool) and fast_id < len(
            self._running_job_pool
        ):
            job_id = self._running_job_pool[fast_id]
            if not job_id:
                fast_id += 1
            else:
                if fast_id != slow_id:
                    self._running_job_pool[slow_id] = job_id
                    self._running_job_pool[fast_id] = None
                fast_id += 1
                slow_id += 1

        # We still need to track the location of current job.
        for i in range(slow_id):
            if self._running_job_pool[i] == current_job_id:
                self._running_job_idx = i
                break

    def update_job_pool_size(self, new_pool_size):
        """
        This is the call exposed to external components.
        [TODO] Spawn a new process to monitor resource and apply
        necessary adjustment from here.
        """
        # This is for the actual resource adjustment.
        assert new_pool_size >= 1
        self._job_pool_size = new_pool_size

        if self._job_pool_size < len(self._running_job_pool):
            self.shrink_resource()

        if self._job_pool_size > len(self._running_job_pool):
            self.grow_resource()

    def shrink_resource(self):
        # Always move jobs to front.
        self.tighten_jobs()

        while (
            len(self._running_job_pool) > self._job_pool_size
            and not self._running_job_pool[-1]
        ):
            del self._running_job_pool[-1]

        # Note that it is still possible that we can not shrink it
        # when there are jobs in progress
        print(
            "Shrink job pool size in JobScheduler!",
            len(self._running_job_pool),
            self._job_pool_size,
        )

    def grow_resource(self):
        # Always move jobs to front.
        self.tighten_jobs()

        # This does not influence the relative order of jobs with self._running_job_idx
        while len(self._running_job_pool) < self._job_pool_size:
            self._running_job_pool.append(None)


class JobScheduler:
    def __init__(
        self,
        job_manager,
        pool_size,
        cc_controller=BasicCongestionControl(DEFAULT_JOB_POOL_SIZE),
        policy=SchedulePolicy.FIFO,
    ):
        """
        self._jobs_queue are all the jobs.
        self._due_jobs_list stores all potential jobs that can be marked
        as expired jobs.
        self._inactive_jobs are jobs that are already invalid.
        """
        self._job_manager = job_manager
        self.interval = EXPIRE_INTERVAL
        self._jobs_queue = queue.Queue()
        self._inactive_jobs = set()
        self._due_jobs_list = []

        self._CC_controller = cc_controller
        self._current_pool_size = self._CC_controller._job_pool_size
        # Start the loop process in an async way
        asyncio.create_task(self.job_cleanup_loop())
        self._policy = policy

    def configure_job_pool_size(self, new_pool_size):
        # Here it just set the pool size, later when it starts scheduling
        # we will update appropriate slots correspondingly.
        self._current_pool_size = new_pool_size

    def append_job(self, job_id, due_time_seconds):
        # This submits a job to scheduler. The scheduler will determine
        # which job gets executed.
        self._jobs_queue.put(job_id)

        def key_func(x):
            return x[1]

        current_time = time.time()
        due_time = current_time + due_time_seconds
        item = (job_id, due_time)
        index = bisect.bisect_left(
            [key_func(t) for t in self._due_jobs_list], key_func(item)
        )
        self._due_jobs_list.insert(index, item)

    def schedule_next_job(self):
        # Scheduler outputs a job to be processed following the specified policy.
        job_id = None

        # [TODO] use class abstraction for SchedulingPolicy
        if self._policy == SchedulePolicy.FIFO:
            if self._jobs_queue.empty():
                print("Job scheduler is waiting jobs coming ......")
                time.sleep(1)
                return job_id
            if not self._jobs_queue.empty():
                job_id = self._jobs_queue.get()

            # Every time when popping a job from queue,
            # we check if this job is in active state.
            while (
                job_id
                and job_id in self._inactive_jobs
                and not self._jobs_queue.empty()
            ):
                job_id = self._jobs_queue.get()

        else:
            print("Unsupported scheduling policy!")

        self._job_manager.start_execute_job(job_id)
        return job_id

    def get_inactive_jobs(self):
        return self._inactive_jobs

    async def expire_jobs(self):
        # This is to expire jobs based on specified due time per job.
        if self._policy == SchedulePolicy.FIFO:
            current_time = time.time()
            idx = 0
            while (
                idx < len(self._due_jobs_list)
                and self._due_jobs_list[idx][1] <= current_time
            ):
                idx += 1

            print("Number of expired jobs is ", idx)
            for i in range(idx):
                # Update job's status to job manager
                job_id = self._due_jobs_list[i][0]
                self._inactive_jobs.add(job_id)
                print("======> ", job_id, "expired.")
            self._due_jobs_list = self._due_jobs_list[idx:]
        else:
            print("Unsupported scheduling policy!")

    async def job_cleanup_loop(self):
        """
        This is a long-running process to check if jobs have expired or not.
        """
        round_id = 0
        while True:
            start_time = time.time()  # Record start time
            await self.expire_jobs()  # Run the process
            elapsed_time = time.time() - start_time  # Calculate elapsed time
            time_to_next_run = max(
                0, self.interval - elapsed_time
            )  # Calculate remaining time
            print("Sliding, round: ", round_id)
            round_id += 1
            await asyncio.sleep(time_to_next_run)  # Wait for the remaining time

    def round_robin_get_job(self):
        # Step 1
        # Before scheduling any new jobs, we need to check the status of previous
        # jobs and update it accordingly.
        for i in range(len(self._CC_controller._running_job_pool)):
            if not self._CC_controller._running_job_pool[i]:
                continue
            job_id = self._CC_controller._running_job_pool[i]
            # Do not schedule new job in since we need to adjust capacity
            # based on new pool size representing how much underlying resource.
            if self._job_manager.get_job_status(job_id) == JobStatus.COMPLETED:
                self._CC_controller._running_job_pool[i] = None

        # Step 2, after the jobs' status are updated,
        # we need to iterate over all slots for next job.
        # By default these jobs' priority is higher than new jobs.
        next_job_id = None
        for i in range(len(self._CC_controller._running_job_pool)):
            temp_idx = self._CC_controller._running_job_idx + 1
            temp_idx = temp_idx % len(self._CC_controller._running_job_pool)
            self._CC_controller._running_job_idx = temp_idx
            job_id = self._CC_controller._running_job_pool[temp_idx]
            if not job_id:
                continue
            else:
                next_job_id = job_id
                break
        if not next_job_id:
            self._CC_controller._running_job_idx = 0

        # Step 3, update job pool size with controller.
        self._CC_controller.update_job_pool_size(self._current_pool_size)

        # Step 4, if there is available slot, schedule new job in from queue.
        start_offset = len(self._CC_controller._running_job_pool)
        for i in range(len(self._CC_controller._running_job_pool)):
            if not self._CC_controller._running_job_pool[i]:
                start_offset = i
                break

        while start_offset < len(self._CC_controller._running_job_pool):
            new_job_id = self.schedule_next_job()
            if not new_job_id:
                break

            if not next_job_id:
                next_job_id = new_job_id
            self._CC_controller._running_job_pool[start_offset] = new_job_id
            start_offset += 1

        if not next_job_id:
            print("No job is found!")

        return next_job_id



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import List, Optional

from huggingface_hub import HfApi, hf_hub_download, snapshot_download

from aibrix import envs
from aibrix.common.errors import (
    ArgNotCongiuredError,
    ArgNotFormatError,
    ModelNotFoundError,
)
from aibrix.downloader.base import (
    DEFAULT_DOWNLOADER_EXTRA_CONFIG,
    BaseDownloader,
    DownloadExtraConfig,
)
from aibrix.logger import init_logger

logger = init_logger(__name__)


def _parse_model_name_from_uri(model_uri: str) -> str:
    return model_uri


class HuggingFaceDownloader(BaseDownloader):
    def __init__(
        self,
        model_uri: str,
        model_name: Optional[str] = None,
        download_extra_config: DownloadExtraConfig = DEFAULT_DOWNLOADER_EXTRA_CONFIG,
        enable_progress_bar: bool = False,
    ):
        if model_name is None:
            model_name = _parse_model_name_from_uri(model_uri)
            logger.info(f"model_name is not set, using `{model_name}` as model_name")

        self.download_extra_config = download_extra_config

        self.hf_token = self.download_extra_config.hf_token or envs.DOWNLOADER_HF_TOKEN
        self.hf_endpoint = (
            self.download_extra_config.hf_endpoint or envs.DOWNLOADER_HF_ENDPOINT
        )
        self.hf_revision = (
            self.download_extra_config.hf_revision or envs.DOWNLOADER_HF_REVISION
        )
        self.hf_api = HfApi(endpoint=self.hf_endpoint, token=self.hf_token)

        super().__init__(
            model_uri=model_uri,
            model_name=model_name,
            bucket_path=model_uri,
            bucket_name=None,
            download_extra_config=download_extra_config,
            enable_progress_bar=enable_progress_bar,
        )  # type: ignore

        # Dependent on the attributes generated in the base class,
        # so place it after the super().__init__() call.
        self.allow_patterns = (
            None
            if self.allow_file_suffix is None or len(self.allow_file_suffix) == 0
            else [f"*.{suffix}" for suffix in self.allow_file_suffix]
        )
        logger.debug(
            f"Downloader {self.__class__.__name__} initialized."
            f"HF Settings are followed: \n"
            f"hf_token={self.hf_token}, \n"
            f"hf_endpoint={self.hf_endpoint}"
        )

    def _valid_config(self):
        if len(self.model_uri.split("/")) != 2:
            raise ArgNotFormatError(arg_name="model_uri", expected_format="repo/name")

        if self.model_name is None:
            raise ArgNotCongiuredError(arg_name="model_name", arg_source="--model-name")

        if not self.hf_api.repo_exists(repo_id=self.model_uri):
            raise ModelNotFoundError(model_uri=self.model_uri)

    def _is_directory(self) -> bool:
        """Check if model_uri is a directory.
        model_uri in `repo/name` format must be a directory.
        """
        return True

    def _directory_list(self, path: str) -> List[str]:
        return self.hf_api.list_repo_files(repo_id=self.model_uri)

    def _support_range_download(self) -> bool:
        return False

    def download(
        self,
        local_path: Path,
        bucket_path: str,
        bucket_name: Optional[str] = None,
        enable_range: bool = True,
    ):
        hf_hub_download(
            repo_id=self.model_uri,
            filename=bucket_path,
            local_dir=local_path,
            revision=self.hf_revision,
            token=self.hf_token,
            endpoint=self.hf_endpoint,
            local_dir_use_symlinks=False,
            force_download=self.force_download,
        )

    def download_directory(self, local_path: Path):
        max_workers = (
            self.download_extra_config.num_threads or envs.DOWNLOADER_NUM_THREADS
        )
        snapshot_download(
            self.model_uri,
            local_dir=local_path,
            revision=self.hf_revision,
            token=self.hf_token,
            allow_patterns=self.allow_patterns,
            max_workers=max_workers,
            endpoint=self.hf_endpoint,
            local_dir_use_symlinks=False,
            force_download=self.force_download,
        )



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from pathlib import Path
from typing import Union

from aibrix import envs
from aibrix.config import DOWNLOAD_CACHE_DIR
from aibrix.logger import init_logger

logger = init_logger(__name__)


def meta_file(local_path: Union[Path, str], file_name: str, source: str) -> Path:
    return (
        Path(local_path)
        .joinpath(DOWNLOAD_CACHE_DIR % source)
        .joinpath(f"{file_name}.metadata")
        .absolute()
    )


def save_meta_data(file_path: Union[Path, str], etag: str):
    Path(file_path).parent.mkdir(parents=True, exist_ok=True)
    with open(file_path, "w") as f:
        f.write(etag)


def load_meta_data(file_path: Union[Path, str]):
    if Path(file_path).exists():
        with open(file_path, "r") as f:
            return f.read()
    return None


def check_file_exist(
    local_file: Union[Path, str],
    meta_file: Union[Path, str],
    expected_file_size: int,
    expected_etag: str,
) -> bool:
    if expected_file_size is None or expected_file_size <= 0:
        return False

    if expected_etag is None or expected_etag == "":
        return False

    if not Path(local_file).exists():
        return False

    file_size = os.path.getsize(local_file)
    if file_size != expected_file_size:
        return False

    if not Path(meta_file).exists():
        return False

    etag = load_meta_data(meta_file)
    return etag == expected_etag


def need_to_download(
    local_file: Union[Path, str],
    meta_data_file: Union[Path, str],
    expected_file_size: int,
    expected_etag: str,
) -> bool:
    _file_name = Path(local_file).name
    if not envs.DOWNLOADER_FORCE_DOWNLOAD and envs.DOWNLOADER_CHECK_FILE_EXIST:
        if check_file_exist(
            local_file, meta_data_file, expected_file_size, expected_etag
        ):
            logger.info(f"File {_file_name} exist in local, skip download.")
            return False
        else:
            logger.info(f"File {_file_name} not exist in local, start to download...")
    else:
        logger.info(
            f"File {_file_name} start downloading directly "
            f"for DOWNLOADER_FORCE_DOWNLOAD={envs.DOWNLOADER_FORCE_DOWNLOAD}, "
            f"DOWNLOADER_CHECK_FILE_EXIST={envs.DOWNLOADER_CHECK_FILE_EXIST}"
        )
    return True


def infer_model_name(uri: str):
    if uri is None or uri == "":
        raise ValueError("Model uri is empty.")

    return uri.strip().strip("/").split("/")[-1]



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import subprocess
import sys  # Import sys for command-line arguments


def main():
    script_path = os.path.join(os.path.dirname(__file__), "benchmark.sh")

    # Get command-line arguments *after* the script name
    args = sys.argv[1:]  # Slice to exclude the Python script name itself

    try:
        subprocess.run([script_path] + args, check=True)  # Pass args to the script
    except subprocess.CalledProcessError as e:
        print(f"Error running benchmark script: {e}")
        exit(1)
    except FileNotFoundError:
        print(f"Error: benchmark.sh not found at {script_path}")
        exit(1)
    except Exception as e:
        print(f"A general error occurred: {e}")
        exit(1)


if __name__ == "__main__":
    main()



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import logging
import math
import os
from datetime import datetime

import numpy as np
import pandas as pd

REDIS_PROFILE_KEY = "aibrix:profile_%s_%s"
TPUT_TOLERANCE = 0.9

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("aibrix.gpu_optimizer.profiling")


def gen(args):
    # Init dataframe and load benchmark results
    benchmark = os.path.dirname(__file__) + f"/result/{args.deployment}.jsonl"
    if args.benchmark is not None:
        benchmark = args.benchmark

    benchmark_results = []
    with open(benchmark, "r") as f:
        for line in f:
            if line == "\n":
                continue
            benchmark_results.append(json.loads(line))
    benchmark_df = pd.DataFrame(
        benchmark_results,
        columns=[
            "input_tokens",
            "output_tokens",
            "request_rate",
            "seed",
            "model",
            "samples",
            "metric",
            "mean",
            "P50",
            "P90",
            "P99",
        ],
    )

    # Construct matrix indexes based on unique input and output tokens
    input_tokens = benchmark_df["input_tokens"].unique()
    output_tokens = benchmark_df["output_tokens"].unique()
    input_tokens.sort()
    output_tokens.sort()
    slo_tputs = np.zeros((len(output_tokens), len(input_tokens)), dtype=float)

    # Create TPAT metric
    tpat_df = benchmark_df[benchmark_df["metric"] == "E2E"].copy()
    tpat_df["metric"] = ["TPAT"] * len(tpat_df)
    for _, percentile in enumerate(["mean", "P50", "P90", "P99"]):
        tpat_df[percentile] = tpat_df[percentile] / (
            tpat_df["input_tokens"] + tpat_df["output_tokens"]
        )
    benchmark_df = pd.concat([benchmark_df, tpat_df], ignore_index=True)

    # Decide the percentile to use for SLO calculation
    percentile_field = "mean"
    if args.percentile > 0:
        percentile_field = f"P{args.percentile}"

    # Iterate slo_tputs and fill in the matrix with the throughput values that matches the SLO
    for i in range(len(output_tokens)):
        for j in range(len(input_tokens)):
            filtered_df = benchmark_df.loc[
                (benchmark_df["input_tokens"] == input_tokens[j])
                & (benchmark_df["output_tokens"] == output_tokens[i])
            ]

            # Filter the bencmarks by throughput SLO. Besides, we arbitarily filter out instable throughput as TPUT < request_rate * TPUT_TOLERANCE
            tput_df = filtered_df.loc[
                (filtered_df["metric"] == "TPUT")
                & (filtered_df["mean"] >= args.tput)
                & (filtered_df["mean"] >= filtered_df["request_rate"] * TPUT_TOLERANCE)
            ]
            if len(tput_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(tput_df["request_rate"])
            ]

            # Filter the bencmarks by token throughput SLO
            tt_df = filtered_df.loc[
                (filtered_df["metric"] == "TT") & (filtered_df["mean"] >= args.tt)
            ]
            if len(tt_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(tt_df["request_rate"])
            ]

            # Filter the bencmarks by E2E latency SLO
            e2e_df = filtered_df.loc[
                (filtered_df["metric"] == "E2E")
                & (filtered_df[percentile_field] <= args.e2e)
            ]
            if len(e2e_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(e2e_df["request_rate"])
            ]

            # Filter the bencmarks by TTFT SLO
            ttft_df = filtered_df.loc[
                (filtered_df["metric"] == "TTFT")
                & (filtered_df[percentile_field] <= args.ttft)
            ]
            if len(ttft_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(ttft_df["request_rate"])
            ]

            # Filter the bencmarks by TPAT SLO
            tpat_df = filtered_df.loc[
                (filtered_df["metric"] == "TPAT")
                & (filtered_df[percentile_field] <= args.tpat)
            ]
            if len(tpat_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(tpat_df["request_rate"])
            ]

            # Filter the bencmarks by TPOT SLO

            tpot_df = filtered_df.loc[
                (filtered_df["metric"] == "TPOT")
                & (filtered_df[percentile_field] <= args.tpot)
            ]
            if len(tpot_df) == 0:
                continue
            filtered_df = filtered_df.loc[
                filtered_df["request_rate"].isin(tpot_df["request_rate"])
            ]

            # Conclude
            slo_tputs[i, j] = np.max(
                filtered_df.loc[filtered_df["metric"] == "TPUT", "mean"]
            )
            msg = filtered_df.loc[
                filtered_df["metric"] == "TPUT", ["request_rate", "mean"]
            ]
            logger.debug(
                f"Candidate for {input_tokens[j]}:{output_tokens[i]} tputs: {msg}"
            )

    # Print the matrix
    result = {
        "gpu": args.deployment,
        "cost": args.cost,
        "tputs": slo_tputs.tolist(),
        "indexes": [output_tokens.tolist(), input_tokens.tolist()],
        "created": datetime.now().timestamp(),
    }
    if args.o is not None:
        if _try_store_redis(args, result):
            return

        with open(args.o, "w") as f:
            json.dump(result, f)
    else:
        print(json.dumps(result))


def _try_store_redis(args, result) -> bool:
    import json
    import sys
    from urllib.parse import parse_qs, urlparse

    import redis

    # Parse the Redis URL
    url = urlparse(args.o)

    # Connect to the Redis server
    if url.scheme != "redis":
        return False

    # Connect to the Redis server
    db_name = str(url.path).strip("/")
    if db_name == "":
        db_name = "0"
    redis_client = redis.Redis(
        host=str(url.hostname),
        port=6379 if url.port is None else int(url.port),
        db=int(db_name),
        username=url.username,
        password=url.password,
    )

    # Store the result in Redis
    query_params = parse_qs(url.query)
    model_name = query_params.get("model", [""])[0]
    if model_name == "":
        print('"model" in Redic connection arguments is not provided.', file=sys.stderr)
        return True

    redis_key = REDIS_PROFILE_KEY % (model_name, args.deployment)
    redis_client.set(redis_key, json.dumps(result))
    print(f"Result stored in Redis: {redis_key}.")
    return True


def main():
    parser = argparse.ArgumentParser(
        description="Benchmark the online serving throughput."
    )
    parser.add_argument(
        "deployment", type=str, help="Target deployment", default="result"
    )
    parser.add_argument(
        "--benchmark", type=str, default=None, help="Benchmark result file."
    )
    parser.add_argument(
        "--tput", type=float, default=0, help="Throughput SLO target as RPS."
    )
    parser.add_argument(
        "--tt", type=float, default=0, help="Token Throughput SLO target."
    )
    parser.add_argument(
        "--e2e", type=float, default=math.inf, help="E2E latency SLO target."
    )
    parser.add_argument(
        "--ttft", type=float, default=math.inf, help="Time To First Token SLO target."
    )
    parser.add_argument(
        "--tpat", type=float, default=math.inf, help="Time Per All Token SLO target."
    )
    parser.add_argument(
        "--tpot", type=float, default=math.inf, help="Time Per Output Token SLO target."
    )
    parser.add_argument(
        "--percentile",
        type=int,
        default=0,
        help="Percentile to use for SLO calculation. Default to ignore percentile and use mean.",
        choices=[0, 50, 90, 99],
    )
    parser.add_argument("--cost", type=float, default=1.0, help="Cost of the GPU.")
    parser.add_argument(
        "-o",
        type=str,
        default=None,
        help="Output file name. support redis as: redis://[username:password@]hostname:port[/db_name]?model=[model_name]",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="Print more information for understanding the generated profile",
    )

    args = parser.parse_args()
    if args.verbose:
        logger.setLevel(level=logging.DEBUG)

    gen(args)


if __name__ == "__main__":
    main()



import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Union

from .solver import MelangeSolver, Solver

PROJECT_DIR = Path(__file__).parent.parent.parent


@dataclass
class Config:
    gpu_info: dict = field(default_factory=dict)
    workload_distribution: list = field(default_factory=list)
    total_request_rate: float = 0  # units: requests per second
    slice_factor: int = 4


# This class is adapted from code originally written by Tyler Griggs
# found at https://github.com/tyler-griggs/melange-release
# See: https://tyler-griggs.github.io/blogs/melange
class SolverRunner:
    def __init__(self, config: Union[str, Config]):
        if isinstance(config, str):
            config = Config(**json.load(open(config)))
        self.config: Config = config
        self.solver: Solver = MelangeSolver(
            workload_distribution=self.config.workload_distribution,
            total_request_rate=self.config.total_request_rate,
            gpu_info=self.config.gpu_info,
            slice_factor=self.config.slice_factor,
        )
        self.execution_result = {}  # type: ignore

    def run(self):
        self.execution_result = self.solver.run()
        return self.execution_result

    def export(self, path):
        with open(path, "w") as f:
            json.dump(self.execution_result, f, indent=4)



import pulp
from pulp import LpInteger, LpMinimize, LpProblem, LpVariable

from .util import tputs_to_loads_2d


# This class is adapted from code originally written by Tyler Griggs
# found at https://github.com/tyler-griggs/melange-release
# See: https://tyler-griggs.github.io/blogs/melange
class Solver:
    def __init__(
        self, workload_distribution: list, total_request_rate: float, gpu_info: dict
    ):
        self.workload_distribution = workload_distribution
        self.overall_rate = total_request_rate
        self.gpu_info = gpu_info

    def run(self, logs=False):
        raise NotImplementedError


class MelangeSolver(Solver):
    def __init__(
        self,
        workload_distribution: list,
        total_request_rate: float,
        gpu_info: dict,
        slice_factor: int,
    ):
        super().__init__(workload_distribution, total_request_rate, gpu_info)
        self.slice_factor = slice_factor

    def run(self, logs=False):
        # Multiply overall rate across distribution.
        request_rate_histogram = []
        for i in range(len(self.workload_distribution)):
            request_rate_histogram.append([])
            for j in range(len(self.workload_distribution[0])):
                request_rate_histogram[-1].append(
                    self.workload_distribution[i][j] * self.overall_rate
                )

        # Convert the profiled max throughputs into mapping from request size to load
        for gpu in self.gpu_info:
            self.gpu_info[gpu]["loads"] = tputs_to_loads_2d(self.gpu_info[gpu]["tputs"])

        gpu_types = list(self.gpu_info.keys())
        cost_vector = [self.gpu_info[gpu]["cost"] for gpu in gpu_types]

        # Create slices, which is a single dimension.
        slices = []
        for i in range(len(request_rate_histogram)):
            for j in range(len(request_rate_histogram[i])):
                for _ in range(self.slice_factor):
                    slices.append(request_rate_histogram[i][j] / self.slice_factor)

        # Create slice-to-load mapping, which is a single dimension.
        for gpu in gpu_types:
            slice_loads = []
            for i in range(len(self.gpu_info[gpu]["loads"])):
                for j in range(len(self.gpu_info[gpu]["loads"][i])):
                    for _ in range(self.slice_factor):
                        slice_loads.append(self.gpu_info[gpu]["loads"][i][j])
            assert len(slices) == len(slice_loads)
            self.gpu_info[gpu]["slice_loads"] = slice_loads

        # Decision matrix value is binary. The slice is assigned to a GPU, or it isn't.
        matrix_rows = len(slices)
        matrix_cols = len(gpu_types)

        # Vector value is non-negative integer of how many of each GPU type are needed
        vector_length = matrix_cols

        decision_matrix = [
            [
                LpVariable(f"x_{i}_{j}", cat=LpInteger, lowBound=0, upBound=1)
                for j in range(matrix_cols)
            ]
            for i in range(matrix_rows)
        ]
        decision_vector = [
            LpVariable(f"y_{i}", cat=LpInteger, lowBound=0)
            for i in range(vector_length)
        ]

        # Objective: minimize cost
        problem = LpProblem("GpuAllocation", LpMinimize)
        problem += pulp.lpSum(
            [decision_vector[i] * cost_vector[i] for i in range(len(decision_vector))]
        )

        # C1: Each row of decision matrix must sum to exactly 1 (ie, each slice assigned to one GPU)
        for i in range(len(decision_matrix)):
            problem += pulp.lpSum(decision_matrix[i]) == 1

        # C2: Load of column of decision matrix must fit in decision vector capacity
        for j in range(len(decision_matrix[0])):
            # j is idx of GPU type, i is slice
            problem += (
                pulp.lpSum(
                    [
                        decision_matrix[i][j]
                        * self.gpu_info[gpu_types[j]]["slice_loads"][i]
                        * slices[i]
                        for i in range(len(decision_matrix))
                    ]
                )
                <= decision_vector[j]
            )

        # Solve the problem
        problem.solve(pulp.PULP_CBC_CMD(msg=0))

        # For Arm-based Mac platforms.
        # solver= pulp.getSolver('COIN_CMD', path='/opt/homebrew/opt/cbc/bin/cbc', msg=0)
        # problem.solve(solver)

        # Print the results if needed
        if logs:
            print("Decision Matrix:")
            for row in decision_matrix:
                print([var.value() for var in row])
            print("Decision Vector:")
            print(f"{[var.value() for var in decision_vector]}")

        if pulp.LpStatus[problem.status] != "Optimal":
            return None

        solution_dict = {}
        for i in range(len(decision_vector)):
            solution_dict[gpu_types[i]] = int(decision_vector[i].value())

        total_cost = 0
        for gpu in solution_dict:
            total_cost += solution_dict[gpu] * self.gpu_info[gpu]["cost"]
        solution_dict["cost"] = total_cost

        return solution_dict



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from abc import ABC, abstractmethod
from enum import Enum

LOCAL_STORAGE_PATH_VAR = "LOCAL_STORAGE_PATH"


class StorageType(Enum):
    LocalDiskFile = 1
    S3 = 2
    TOS = 3


class PersistentStorage(ABC):
    """
    This is an abstract class.

    A storage should implement this class, such as Local files, TOS and S3.
    Any storage implementation are transparent to external components.
    """

    @abstractmethod
    def write_job_input_data(self, job_id, inputDataFileName):
        pass

    @abstractmethod
    def read_job_input_data(self, job_id, start_index=0, num_requests=-1):
        pass

    @abstractmethod
    def write_job_output_data(self, job_id, output_list):
        pass

    @abstractmethod
    def delete_job_data(self, job_id):
        pass

    def get_current_storage_type(self):
        print("Current storage type: ", self._storage_type_name)
        return self._storage_type_name

    @classmethod
    def create_storage(cls, storage_type=StorageType.LocalDiskFile):
        if storage_type == StorageType.LocalDiskFile:
            return LocalDiskFiles()
        else:
            raise ValueError("Unknown storage type")


class LocalDiskFiles(PersistentStorage):
    """
    This stores all job data in local disk as files.
    """

    def __init__(self):
        print("Setup ENV VAR for local storage path! ")

        if LOCAL_STORAGE_PATH_VAR in os.environ:
            self.directory_path = os.environ[LOCAL_STORAGE_PATH_VAR]
        else:
            self.directory_path = os.path.abspath(os.path.dirname(__file__))

        self.directory_path = self.directory_path + "/data/"
        os.makedirs(self.directory_path, exist_ok=True)
        print("Storage path is located: ", self.directory_path)

    def write_job_input_data(self, job_id, inputDataFileName):
        """This writes requests file to local disk."""
        request_list = []
        # Open the JSON file
        with open(inputDataFileName, "r") as file:
            # Parse JSON data into a Python dictionary
            for line in file.readlines():
                if len(line) <= 1:
                    continue
                data = json.loads(line)
                request_list.append(data)

        num_valid_request = len(request_list)
        print(f"Storage side receives {num_valid_request} request.")

        directory_path = self.directory_path + str(job_id) + "/"
        os.makedirs(directory_path)

        inputFileName = "input.json"
        inputJsonName = directory_path + inputFileName
        with open(inputJsonName, "w") as file:
            for obj in request_list:
                file.write(json.dumps(obj) + "\n")

    def read_job_input_data(self, job_id, start_index=0, num_requests=-1):
        """Read job requests input from local disk."""
        directory_path = self.directory_path + str(job_id) + "/"
        inputFileName = "input.json"
        inputJsonName = directory_path + inputFileName

        request_inputs = []
        if not os.path.exists(inputJsonName):
            print(f"job {job_id} does not exist in storage!")
            return request_inputs

        with open(inputJsonName, "r") as file:
            for _ in range(start_index):
                next(file)
                if not file:
                    print("read requests is out of index, not enough size.")
                    return request_inputs

            if num_requests > 0:
                for _ in range(num_requests):
                    line = file.readline()
                    if not line:  # End of file reached
                        break
                    data = json.loads(line)
                    request_inputs.append(data)
            else:
                # Parse JSON data into a Python dictionary
                for line in file.readlines():
                    if len(line) <= 1:
                        continue
                    data = json.loads(line)
                    request_inputs.append(data)
        # print("debug: ", len(request_inputs))
        return request_inputs

    def get_job_number_requests(self, job_id):
        """Get job requests length from local disk."""
        directory_path = self.directory_path + str(job_id) + "/"
        inputFileName = "input.json"
        inputJsonName = directory_path + inputFileName

        if not os.path.exists(inputJsonName):
            print(f"job {job_id} does not exist in storage!")
            return 0

        with open(inputJsonName, "r") as file:
            return sum(1 for line in file)

        return 0

    def write_job_output_data(self, job_id, start_index, output_list):
        """Write job results output as files."""
        directory_path = self.directory_path + str(job_id) + "/"

        if not os.path.exists(directory_path):
            print(
                f"Error: Job {job_id} does not exist, perhaps need to create Job first!"
            )
            return
        output_file_path = directory_path + "output.json"

        with open(output_file_path, "a+") as file:
            for _ in range(start_index):
                next(file, None)
                if not file:
                    print("writing requests is out of index.")
                    return

            for obj in output_list:
                file.write(json.dumps(obj) + "\n")
            file.truncate()

    def read_job_output_data(self, job_id, start_index, num_requests):
        """Read job results output from local disk as files."""
        directory_path = self.directory_path + str(job_id) + "/"

        output_data = []
        if not os.path.exists(directory_path):
            print(
                f"Error: Job {job_id} does not exist, perhaps need to create Job first!"
            )
            return output_data
        output_file_path = directory_path + "output.json"

        with open(output_file_path, "r") as file:
            for _ in range(start_index):
                next(file)
                if not file:
                    print("reading requests output is out of index.")
                    return output_data

            num_lines = 0
            for line in file.readlines():
                if len(line) <= 1:
                    continue
                data = json.loads(line)
                output_data.append(data)
                num_lines += 1
                if num_lines == num_requests:
                    break

        return output_data

    def delete_job_data(self, job_id):
        """Delete all input and output files for the job."""
        directory_path = self.directory_path + str(job_id) + "/"

        input_file_path = directory_path + "input.json"
        try:
            os.remove(input_file_path)
        except FileNotFoundError:
            print(f"Job ID {input_file_path} does not exist.")
        except Exception as e:
            print(f"Error: {e}")

        output_file_path = directory_path + "output.json"
        if os.path.exists(output_file_path):
            try:
                os.remove(output_file_path)
            except FileNotFoundError:
                print(f"Job output {output_file_path} does not exist.")
            except Exception as e:
                print(f"Error: {e}")

        try:
            os.rmdir(directory_path)
        except FileNotFoundError:
            print(f"Job Directory {directory_path} does not exist.")
        except OSError as e:
            print(f"Error: {e} - Job Directory is not empty or can't be deleted.")



import argparse
import os
import shutil
import time
from pathlib import Path
from typing import Optional
from urllib.parse import urljoin

import uvicorn
from fastapi import APIRouter, FastAPI, Request, Response
from fastapi.datastructures import State
from fastapi.responses import JSONResponse
from httpx import Headers
from prometheus_client import make_asgi_app, multiprocess
from starlette.routing import Mount

from aibrix import __version__, envs
from aibrix.config import EXCLUDE_METRICS_HTTP_ENDPOINTS
from aibrix.logger import init_logger
from aibrix.metrics.engine_rules import get_metric_standard_rules
from aibrix.metrics.http_collector import HTTPCollector
from aibrix.metrics.metrics import (
    HTTP_COUNTER_METRICS,
    HTTP_LATENCY_METRICS,
    INFO_METRICS,
    REGISTRY,
)
from aibrix.openapi.engine.base import InferenceEngine, get_inference_engine
from aibrix.openapi.model import ModelManager
from aibrix.openapi.protocol import (
    DownloadModelRequest,
    ErrorResponse,
    ListModelRequest,
    LoadLoraAdapterRequest,
    UnloadLoraAdapterRequest,
)

logger = init_logger(__name__)
router = APIRouter()


def initial_prometheus_multiproc_dir():
    if "PROMETHEUS_MULTIPROC_DIR" not in os.environ:
        prometheus_multiproc_dir = envs.PROMETHEUS_MULTIPROC_DIR
    else:
        prometheus_multiproc_dir = os.environ["PROMETHEUS_MULTIPROC_DIR"]

    # Note: ensure it will be automatically cleaned up upon exit.
    path = Path(prometheus_multiproc_dir)
    path.mkdir(parents=True, exist_ok=True)
    if path.is_dir():
        for item in path.iterdir():
            if item.is_dir():
                shutil.rmtree(item)
            else:
                item.unlink()
    os.environ["PROMETHEUS_MULTIPROC_DIR"] = envs.PROMETHEUS_MULTIPROC_DIR


# filter_and_set_headers only keeps necessary headers
def filter_and_set_headers(request_headers):
    # Create a new Headers object to ensure proper case-insensitive handling
    filtered_headers = Headers()

    # List of headers to retain
    headers_to_keep = ["content-type", "authorization"]

    # the httpx library internally normalizes all header keys to lower case for consistent access and comparison.
    # However, we lowercase the headers for comparison just in case we may switch libraries.
    for header, value in request_headers.items():
        if header.lower() in headers_to_keep:
            filtered_headers[header] = value

    return filtered_headers


def inference_engine(request: Request) -> InferenceEngine:
    # header are dynamic for each request, allocate headers in runtime
    engine = request.app.state.inference_engine
    engine.headers = filter_and_set_headers(request.headers)
    return engine


def mount_metrics(app: FastAPI):
    # setup multiprocess collector
    initial_prometheus_multiproc_dir()
    prometheus_multiproc_dir_path = os.environ["PROMETHEUS_MULTIPROC_DIR"]
    logger.info(
        f"AIBrix to use {prometheus_multiproc_dir_path} as PROMETHEUS_MULTIPROC_DIR"
    )
    # registry = CollectorRegistry()
    multiprocess.MultiProcessCollector(REGISTRY)

    # construct scrape metric config
    engine = envs.INFERENCE_ENGINE

    scrape_endpoint = urljoin(envs.INFERENCE_ENGINE_ENDPOINT, envs.METRIC_SCRAPE_PATH)
    collector = HTTPCollector(scrape_endpoint, get_metric_standard_rules(engine))
    REGISTRY.register(collector)
    logger.info(
        f"AIBrix to scrape metrics from {scrape_endpoint}, use {engine} standard rules"
    )

    # Add prometheus asgi middleware to route /metrics requests
    metrics_route = Mount("/metrics", make_asgi_app(registry=REGISTRY))

    app.routes.append(metrics_route)


def init_app_state(state: State) -> None:
    state.inference_engine = get_inference_engine(
        envs.INFERENCE_ENGINE,
        envs.INFERENCE_ENGINE_VERSION,
        envs.INFERENCE_ENGINE_ENDPOINT,
    )


def inference_engine_ready() -> bool:
    try:
        # Check if the engine is initialized
        # Seems no need to check engine's status since main container has its own checks.
        return (
            True if envs.INFERENCE_ENGINE and envs.INFERENCE_ENGINE_ENDPOINT else False
        )
    except Exception as e:
        logger.error(f"Readiness check failed: {e}")
        return False


@router.post("/v1/lora_adapter/load")
async def load_lora_adapter(request: LoadLoraAdapterRequest, raw_request: Request):
    response = await inference_engine(raw_request).load_lora_adapter(request)
    if isinstance(response, ErrorResponse):
        return JSONResponse(content=response.model_dump(), status_code=response.code)

    return Response(status_code=200, content=response)


@router.post("/v1/lora_adapter/unload")
async def unload_lora_adapter(request: UnloadLoraAdapterRequest, raw_request: Request):
    response = await inference_engine(raw_request).unload_lora_adapter(request)
    if isinstance(response, ErrorResponse):
        return JSONResponse(content=response.model_dump(), status_code=response.code)

    return Response(status_code=200, content=response)


# /v1/models is a query to inference engine, this is different from following
# /v1/model/list which is used to fetch runtime managed models locally.
@router.get("/v1/models")
async def list_engine_models(raw_request: Request):
    response = await inference_engine(raw_request).list_models()
    if isinstance(response, ErrorResponse):
        return JSONResponse(content=response.model_dump(), status_code=response.code)

    return JSONResponse(status_code=200, content=response)


@router.post("/v1/model/download")
async def download_model(request: DownloadModelRequest):
    response = await ModelManager.model_download(request)
    if isinstance(response, ErrorResponse):
        return JSONResponse(content=response.model_dump(), status_code=response.code)

    return JSONResponse(status_code=200, content=response.model_dump())


@router.get("/v1/model/list")
async def list_model(request: Optional[ListModelRequest] = None):
    response = await ModelManager.model_list(request)
    if isinstance(response, ErrorResponse):
        return JSONResponse(content=response.model_dump(), status_code=response.code)

    return JSONResponse(status_code=200, content=response.model_dump())


@router.get("/healthz")
async def liveness_check():
    # Simply return a 200 status for liveness check
    return JSONResponse(content={"status": "ok"}, status_code=200)


@router.get("/ready")
async def readiness_check():
    # Check if the inference engine is ready
    if inference_engine_ready():
        return JSONResponse(content={"status": "ready"}, status_code=200)
    return JSONResponse(content={"status": "not ready"}, status_code=503)


def build_app(args: argparse.Namespace):
    if args.enable_fastapi_docs:
        app = FastAPI(debug=False)
    else:
        app = FastAPI(debug=False, openapi_url=None, docs_url=None, redoc_url=None)

    INFO_METRICS.info(
        {
            "version": __version__.__version__,
            "engine": envs.INFERENCE_ENGINE,
            "engine_version": envs.INFERENCE_ENGINE_VERSION,
        }
    )
    mount_metrics(app)
    init_app_state(app.state)
    app.include_router(router)

    @app.middleware("http")
    async def add_router_prometheus_middlerware(request: Request, call_next):
        method = request.method
        endpoint = request.scope.get("path")
        # Exclude endpoints that do not require metrics
        if endpoint in EXCLUDE_METRICS_HTTP_ENDPOINTS:
            response = await call_next(request)
            return response

        start_time = time.perf_counter()
        response = await call_next(request)
        process_time = time.perf_counter() - start_time

        status = response.status_code
        HTTP_LATENCY_METRICS.labels(
            method=method, endpoint=endpoint, status=status
        ).observe(process_time)
        HTTP_COUNTER_METRICS.labels(
            method=method, endpoint=endpoint, status=status
        ).inc()
        return response

    return app


def nullable_str(val: str):
    if not val or val == "None":
        return None
    return val


def main():
    parser = argparse.ArgumentParser(description="Run aibrix runtime server")
    parser.add_argument("--host", type=nullable_str, default=None, help="host name")
    parser.add_argument("--port", type=int, default=8080, help="port number")
    parser.add_argument(
        "--enable-fastapi-docs",
        action="store_true",
        default=False,
        help="Enable FastAPI's OpenAPI schema, Swagger UI, and ReDoc endpoint",
    )
    args = parser.parse_args()
    logger.info("Use %s to startup runtime server", args)
    app = build_app(args=args)
    uvicorn.run(app, host=args.host, port=args.port)


if __name__ == "__main__":
    main()



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import math
import sys
from datetime import datetime
from typing import Callable, Iterable, List, Optional, Protocol, Tuple, Union

from incdbscan import IncrementalDBSCAN

from aibrix.gpu_optimizer.utils import DelayedLog

from .helpers import Centeroid, DataPoints

logger = logging.getLogger("aibrix.gpuoptimizer.clusterer")


class Clusterer(Protocol):
    def insert(self, points):
        """Pass in a list of data points to be clustered"""

    def reset(self):
        """Reset the clusterer"""

    def get_cluster_labels(
        self, points: DataPoints, uncategorized: Optional[List] = None
    ) -> Tuple[Iterable[int], Iterable[Centeroid]]:
        """Get the cluster labels for the given data points"""

    @property
    def length(self):
        """Get the number of data points in the clusterer"""

    @property
    def window(self):
        """Get the time window the data points covers"""


class DBSCANClusterer:
    def __init__(self, eps: float, min_pts: int):
        self.eps = eps
        self.min_pts = min_pts
        self._pt_start: float = math.inf
        self._pt_end: float = 1.0
        self.reset()

    def insert(self, points: DataPoints):
        self.clusterer.insert(points.signatures)
        self._length += len(points)
        ages = tuple(points.datapoint(i).age for i in range(len(points)))
        self._pt_start = min(self._pt_start, *ages)
        self._pt_end = max(self._pt_end, *ages)

    def reset(self):
        self.clusterer = IncrementalDBSCAN(eps=self.eps, min_pts=self.min_pts)
        self._pt_start = math.inf
        self._pt_end = 1.0
        self._length = 0
        self.created = datetime.now().timestamp()

    def clone(self):
        return DBSCANClusterer(self.eps, self.min_pts)

    def get_cluster_labels(
        self, points: DataPoints, uncategorized: Optional[List] = None
    ) -> Tuple[Iterable[int], Iterable[Centeroid]]:
        labels = self.clusterer.get_cluster_labels(points.signatures)
        centers = {}
        start_label = sys.maxsize
        for i, label in enumerate(labels):
            if math.isnan(label):
                continue
            if label < 0:
                if uncategorized is not None:
                    uncategorized.append(points.datapoint(i))
                continue
            start_label = min(start_label, label)
            if label not in centers:
                centers[label] = Centeroid(span=self.window)
            if len(centers) > 10:
                print(f"unepxected label:{label}")
            centers[label].add(points.datapoint(i))
        # Try fixing label index.
        if start_label == sys.maxsize:
            start_label = 0
        if start_label != 0:
            for i, label in enumerate(labels):
                if label >= 0:
                    labels[i] -= start_label
        return labels, centers.values()

    @property
    def length(self):
        return self._length

    @property
    def window(self):
        return (
            1.0
            if self._pt_end == 0.0 or self._pt_end == self._pt_start
            else self._pt_end - self._pt_start
        )


class MovingDBSCANClusterer:
    """MovingCluster uses extra buffer space to store a moving DBSCAN cluster"""

    def __init__(
        self,
        eps: float,
        min_pts: int,
        buffer_size=4,
        window: Union[int, float, Callable[[DBSCANClusterer], bool]] = 4000,
    ):
        if isinstance(window, int):
            self.window_cb = self._get_points_window_cb(window)
        elif isinstance(window, float):
            self.window_cb = self._get_time_window_cb(window)
        else:
            self.window_cb = window

        self.buffer_size = buffer_size
        self.frontier = 0
        self.clusterers = [DBSCANClusterer(eps, min_pts)]
        self._reason = None

    def validate(self) -> bool:
        """Do necessary window rotating and return if data refreshing is necessary"""
        current = (self.frontier + 1) % len(self.clusterers)
        if not self.window_cb(self.clusterers[self.frontier]):
            return False

        # Reset next slot in buffer.
        if len(self.clusterers) < self.buffer_size:
            self.clusterers.append(self.clusterers[current].clone())
            self.frontier = len(self.clusterers) - 1
            logger.debug(
                "moving buffer created: %s, buffers: %s",
                self._reason,
                DelayedLog(lambda: [cluster.length for cluster in self.clusterers]),
            )
            return False
        else:
            self.clusterers[current].reset()
            self.frontier = current
            current = (current + 1) % self.buffer_size
            logger.debug(
                "moving buffer created: %s, now available: %s, buffers: %s",
                self._reason,
                self.clusterer.length,
                DelayedLog(lambda: [cluster.length for cluster in self.clusterers]),
            )
            return True
            # data.trim_head(-cluster['clusterers'][0][current].length)

    def insert(self, points: DataPoints):
        for clusterer in self.clusterers:
            clusterer.insert(points)

    def reset(self):
        self.clusterers = [self.clusterers[0].clone()]
        self.frontier = 0

    def get_cluster_labels(
        self, points: DataPoints, uncategorized: Optional[List] = None
    ) -> Tuple[Iterable[int], Iterable[Centeroid]]:
        return self.clusterer.get_cluster_labels(points, uncategorized=uncategorized)

    @property
    def length(self):
        return self.clusterer.length

    @property
    def window(self):
        return self.clusterer.window

    @property
    def clusterer(self):
        return self.clusterers[(self.frontier + 1) % len(self.clusterers)]

    def _get_points_window_cb(self, window: int) -> Callable[[DBSCANClusterer], bool]:
        return (
            lambda clusterer: clusterer.length >= window / self.buffer_size
            and self.reason(f"reached {round(window / self.buffer_size)} points")
        )

    def _get_time_window_cb(self, window: float) -> Callable[[DBSCANClusterer], bool]:
        return (
            lambda clusterer: datetime.now().timestamp() - clusterer.created
            >= window / self.buffer_size
            and self.reason(
                f"timeout after {round(window / self.buffer_size, 2)} seconds"
            )
        )

    def reason(self, msg) -> bool:
        """Provide a reason for the window to be rotated, always return True."""
        self._reason = msg
        return True



from typing import List


# Convert max throughput profiling to a mapping from request size to load
def tputs_to_loads_2d(max_tputs: List[List[float]]):
    loads: List[List[float]] = []
    for i in range(len(max_tputs)):
        loads.append([])
        for j in range(len(max_tputs[0])):
            load = 1000000.0  # inf
            if max_tputs[i][j] > 0:
                load = 1.0 / max_tputs[i][j]
            loads[-1].append(load)
    return loads



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# The following are all constants.
# This is the time interval for the sliding window to check.
EXPIRE_INTERVAL = 1
# This is the job pool size in job scheduler.
# It should be proportional to resource size in the backend.
DEFAULT_JOB_POOL_SIZE = 1



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field


class NoExtraBaseModel(BaseModel):
    # The class does not allow extra fields
    model_config = ConfigDict(extra="forbid")


class NoProtectedBaseModel(BaseModel):
    # The class does not allow extra fields
    model_config = ConfigDict(extra="forbid", protected_namespaces=())


class ErrorResponse(NoExtraBaseModel):
    object: str = "error"
    message: str
    type: str
    param: Optional[str] = None
    code: int


class LoadLoraAdapterRequest(NoExtraBaseModel):
    lora_name: str
    lora_path: str


class UnloadLoraAdapterRequest(NoExtraBaseModel):
    lora_name: str
    lora_int_id: Optional[int] = Field(default=None)


class DownloadModelRequest(NoProtectedBaseModel):
    model_uri: str
    local_dir: Optional[str] = None
    model_name: Optional[str] = None
    download_extra_config: Optional[Dict] = None


class ModelStatusCard(NoProtectedBaseModel):
    model_name: str
    model_root_path: str
    source: str
    model_status: str


class ListModelRequest(NoExtraBaseModel):
    local_dir: str


class ListModelResponse(NoExtraBaseModel):
    object: str = "list"
    data: List[ModelStatusCard] = Field(default_factory=list)



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from abc import abstractmethod
from typing import Iterable

from prometheus_client import Metric
from prometheus_client.samples import Sample


class StandardRule:
    def __init__(self, rule_type):
        self.rule_type = rule_type

    @abstractmethod
    def __call__(self, metric: Metric) -> Iterable[Metric]:
        pass


class RenameStandardRule(StandardRule):
    def __init__(self, original_name, new_name):
        super().__init__("RENAME")
        self.original_name = original_name
        self.new_name = new_name

    def __call__(self, metric: Metric) -> Iterable[Metric]:
        assert (
            metric.name == self.original_name
        ), f"Metric name {metric.name} does not match Rule original name {self.original_name}"
        metric.name = self.new_name

        # rename all the samples
        _samples = []
        for s in metric.samples:
            s_name = self.new_name + s.name[len(self.original_name) :]
            _samples.append(
                Sample(
                    s_name,
                    s.labels,
                    s.value,
                    s.timestamp,
                    s.exemplar,
                )
            )
        metric.samples = _samples
        yield metric



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

import numpy as np

from aibrix.gpu_optimizer.load_monitor.load_reader import (
    DatasetLoadReader,
    unittest_filepath,
)


class TestDatasetLoadReader(unittest.TestCase):
    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)

        self.reader = DatasetLoadReader(unittest_filepath)

    def test_stair_agggregate(self):
        series = np.array(
            [7, 8, 59, 127, 128, 341, 1023, 1024, 2047, 2048, 3100, 4100, 9000, 10150],
            dtype=float,
        )
        expected = np.array(
            [1, 8, 56, 120, 128, 320, 960, 1024, 1984, 2048, 3072, 4096, 8192, 9216],
            dtype=float,
        )
        np.testing.assert_array_equal(
            self.reader.stair_aggregate(series, skip_log2=True), expected
        )


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


DEFAULT_METRIC_COLLECTOR_TIMEOUT = 1
"""
DOWNLOAD CACHE DIR would be like:
.
└── .cache
    └── huggingface | s3 | tos
        ├── .gitignore
        └── download
"""
DOWNLOAD_CACHE_DIR = ".cache/%s/download"
DOWNLOAD_FILE_LOCK_CHECK_TIMEOUT = 10

EXCLUDE_METRICS_HTTP_ENDPOINTS = ["/metrics/"]



# Copyright 2024 The Aibrix Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from logging import Logger
from logging.handlers import RotatingFileHandler
from pathlib import Path


def _default_logging_basic_config() -> None:
    Path("/tmp/aibrix").mkdir(parents=True, exist_ok=True)
    logging.basicConfig(
        format="%(asctime)s - %(filename)s:%(lineno)d - %(funcName)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S %Z",
        handlers=[
            logging.StreamHandler(stream=sys.stdout),
            RotatingFileHandler(
                "/tmp/aibrix/python.log",
                maxBytes=10 * (2**20),
                backupCount=10,
            ),
        ],
        level=logging.INFO,
    )


def init_logger(name: str) -> Logger:
    return logging.getLogger(name)


_default_logging_basic_config()
logger = init_logger(__name__)
